"""NGPVAN Bulk Import Endpoints"""

import csv
import logging
import uuid

from parsons.etl.table import Table
from parsons.utilities import cloud_storage

logger = logging.getLogger(__name__)


class BulkImport(object):
    def __init__(self):
        pass

    def get_bulk_import_resources(self):
        """
        Get bulk import resources that available to the user. These define
        the types of bulk imports that you can run. These might include
        ``Contacts``, ``ActivistCodes``, ``ContactsActivistCodes`` and others.

        `Returns:`
            list
                A list of resources.
        """

        r = self.connection.get_request("bulkImportJobs/resources")
        logger.info(f"Found {len(r)} bulk import resources.")
        return r

    def get_bulk_import_job(self, job_id):
        """
        Get a bulk import job status.

        `Args:`
            job_id : int
                The bulk import job id.
        `Returns:`
            dict
                The bulk import job
        """

        r = self.connection.get_request(f"bulkImportJobs/{job_id}")
        logger.info(f"Found bulk import job {job_id}.")
        return r

    def get_bulk_import_job_results(self, job_id):
        """
        Get result file of a bulk upload job. This will include one row
        per record processed as well as the status of each.

        If the job results have not been generated, either due to an error in the
        process or the fact the job is still processing, it will return ``None``.

        `Args:`
            job_id: int
                The bulk import job id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.get_bulk_import_job(job_id)
        logger.info(f"Bulk Import Job Status: {r['status']}")
        if r["status"] == "Completed":
            return Table.from_csv(r["resultFiles"][0]["url"])

        return None

    def get_bulk_import_mapping_types(self):
        """
        Get bulk import mapping types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("bulkImportMappingTypes"))
        logger.info(f"Found {tbl.num_rows} bulk import mapping types.")
        return tbl

    def get_bulk_import_mapping_type(self, type_name):
        """
        Get a single bulk import mapping type.

        `Args:`
            type_name: str
        `Returns`:
            dict
                A mapping type json
        """

        r = self.connection.get_request(f"bulkImportMappingTypes/{type_name}")
        logger.info(f"Found {type_name} bulk import mapping type.")
        return r

    def get_bulk_import_mapping_type_fields(self, type_name, field_name):
        """
        Get data about a field in a mapping type.

        `Args:`
            type_name: str
                The mapping type name
            field_name: str
                The field name
        `Returns:`
            dict
                A mapping type fields json
        """

        r = self.connection.get_request(f"bulkImportMappingTypes/{type_name}/{field_name}/values")
        logger.info(f"Found {type_name} bulk import mapping type field values.")
        return r

    def post_bulk_import(
        self,
        tbl,
        url_type,
        resource_type,
        mapping_types,
        description,
        result_fields=None,
        **url_kwargs,
    ):
        # Internal method to post bulk imports.

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(
            tbl,
            url_type,
            file_path=file_name + ".zip",
            quoting=csv.QUOTE_ALL,
            **url_kwargs,
        )
        logger.info(f"Table uploaded to {url_type}.")

        # Generate request json
        json = {
            "description": description,
            "file": {
                "columnDelimiter": "csv",
                "columns": [{"name": c} for c in tbl.columns],
                "fileName": file_name + ".csv",
                "hasHeader": "True",
                "hasQuotes": "True",
                "sourceUrl": url,
            },
            "actions": [
                {
                    "resultFileSizeKbLimit": 5000,
                    "resourceType": resource_type,
                    "actionType": "loadMappedFile",
                    "mappingTypes": mapping_types,
                }
            ],
        }

        if result_fields:
            result_fields = [{"name": c} for c in result_fields]
            json["actions"][0]["columnsToIncludeInResultsFile"] = result_fields

        r = self.connection.post_request("bulkImportJobs", json=json)
        logger.info(f"Bulk upload {r['jobId']} created.")
        return r["jobId"]

    def bulk_apply_activist_codes(self, tbl, url_type, **url_kwargs):
        """
        Bulk apply activist codes.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 50
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ``activistcodeid``
              - Yes
              - A valid activist code id
            * - ``datecanvassed``
              - No
              - An ISO formatted date
            * - ``canvassedby``
              - No
              - A valid User ID; Required when DateCanvassed is provided
            * - ``contacttypeid``
              - No
              - The method of contact.

        `Args:`
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        return self.post_bulk_import(
            tbl,
            url_type,
            "ContactsActivistCodes",
            [{"name": "ActivistCode"}],
            "Activist Code Upload",
            **url_kwargs,
        )

    def bulk_upsert_contacts(self, tbl, url_type, result_fields=None, **url_kwargs):
        """
        Bulk create or update contact records. Provide a Parsons table of contact data to
        create or update records.

        .. note::
            * The first column of the table must be VANID.
            * The other columns can be a combination of the columns listed below.
              The valid column names also accept permutations with underscores, spaces
              and capitalization (e.g. ``phonenumber`` = ``Phone_Number``).

        **Table Fields**

        .. list-table::
            :widths: 500 100 10
            :header-rows: 1

            * - Column
              - Valid Column Names
              - Notes
            * - VANID
              - ``vanid``
              -
            * - Voter VAN ID
              - ``votervanid``
              - The contact's MyVoters VANID
            * - External ID
              - ``externalid``, ``id``, ``pk``, ``voterbaseid``
              - An external id to be stored.
            * - **PII**
              -
              -
            * - First Name
              - ``fn``, ``firstname``, ``first``
              -
            * - Middle Name
              - ``mn``, ``middlename``, ``middle``
              -
            * - Last Name
              - ``ln``, ``lastname``, ``last``
              -
            * - Date of Birth
              - ``dob``, ``dateofbirth``, ``birthdate``
              - An ISO formatted date
            * - Sex
              - ``sex``, ``gender``
              -
            * - **Physical Address**
              -
              -
            * - Address Line 1
              - ``addressline1``, ``address1``, ``address``
              -
            * - Address Line 2
              - ``addressline2``, ``address2``
              -
            * - Address Line 3
              - ``addressline3``, ``address3``
              -
            * - City
              - ``city``
              -
            * - State Or Province
              - ``state``, ``st``, ``stateorprovince``
              -
            * - Zip or Postal Code
              - ``ziporpostal``, ``postal``, ``postalcode``, ``zip``, ``zipcode``
              -
            * - Country Code
              - ``countrycode``, ``country``
              - A valid two character country code (e.g. ``US``)
            * - Display As Entered
              - ``displayasentered``
              - Required values are ``Y`` and ``N``. Determines if the address is
                processed through address correction.
            * - **Phones**
              -
              -
            * - Cell Phone
              - ``cellphone``, ``cell``
              -
            * - Cell Phone Country Code
              - ``cellcountrycode``, ``cellphonecountrycode``
              - A valid two digit country code (e.g. ``01``)
            * - Home Phone
              - ``homephone``, ``home``, ``phone``
              -
            * - Home Phone Country Code
              - ``homecountrycode``, ``homephonecountrycode``
              -
            * - **Email**
              -
              -
            * - Email
              - ``email``, ``emailaddress``
              -
            * - Other Email
              - ``otheremail``, ``email2``, ``emailaddress2``
              -

        `Args:`
            table: Parsons table
              A Parsons table.
            url_type: str
              The cloud file storage to use to post the file. Currently only ``S3``.
            results_fields: list
              A list of fields to include in the results file.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        tbl = tbl.map_columns(CONTACTS_COLUMN_MAP, exact_match=False)

        return self.post_bulk_import(
            tbl,
            url_type,
            "Contacts",
            [{"name": "CreateOrUpdateContact"}],
            "Create Or Update Contact Records",
            result_fields=result_fields,
            **url_kwargs,
        )

    def bulk_apply_suppressions(self, tbl, url_type, **url_kwargs):
        """
        Bulk apply contact suppression codes.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 50
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ``suppressionid``
              - Yes
              - A valid suppression id

        `Args:`
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        return self.post_bulk_import(
            tbl,
            url_type,
            "Contacts",
            [{"name": "Suppressions"}],
            "Apply Suppressions",
            **url_kwargs,
        )

    def bulk_apply_canvass_results(self, tbl, url_type, **url_kwargs):
        """
        Bulk apply contact canvass results.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 50
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ``contacttypeid``
              - Yes
              - Valid Contact Type ID
            * - ``resultid``
              - Yes
              - Valid Contact Result ID
            * - ``datecanvassed``
              - Yes
              - ISO Date Format
            * - ``canvassedby``
              - Yes
              - Valid User ID
            * - ``phone``
              - No
              - Attempted Phone Number
            * - ``countrycode``
              - No
              - Country Code (ISO 3166-1 alpha-2)
            * - ``phonetypeid``
              - No
              - Phone Type
            * - ``phoneoptinstatusid``
              - No
              - SMS Opt-In Status
            * - ``addressid``
              - No
              - The Contact Address ID of the address that was canvassed

        `Args:`
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        return self.post_bulk_import(
            tbl,
            url_type,
            "Contacts",
            [{"name": "CanvassResults"}],
            "Apply Canvass Results",
            **url_kwargs,
        )

    def bulk_apply_contact_custom_fields(self, custom_field_group_id, tbl, url_type, **url_kwargs):
        """
        Bulk apply contact custom fields.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 60
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ***``CF{CustomFieldID}``
              - Yes
              - At least one custom field column to be loaded associated with the provided
                custom_field_group_id. The column name should be a valid Custom Field ID
                prefixed with ``CF``, i.e. CF123.

        `Args:`
            custom_field_group_id: int
                Valid Custom Contact Field Group ID; must be the parent of
                the provided Custom Field IDs in the file.
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        mapping_types = [
            {
                "name": "ApplyContactCustomFields",
                "fieldValueMappings": [
                    {
                        "fieldName": "CustomFieldGroupID",
                        "staticValue": custom_field_group_id,
                    },
                ],
            }
        ]

        return self.post_bulk_import(
            tbl,
            url_type,
            "Contacts",
            mapping_types,
            "Apply Contact Custom Fields",
            **url_kwargs,
        )


# This is a column mapper that is used to accept additional column names and provide
# flexibility for the user.

CONTACTS_COLUMN_MAP = {
    "firstname": ["fn", "first"],
    "middlename": ["mn", "middle"],
    "lastname": ["ln", "last"],
    "dob": ["dateofbirth", "birthdate"],
    "sex": ["gender"],
    "addressline1": ["address", "addressline1", "address1"],
    "addressline2": ["addressline2", "address2"],
    "addressline3": ["addressline3", "address3"],
    "city": [],
    "stateorprovince": ["state", "st"],
    "ziporpostal": ["postal", "postalcode", "zip", "zipcode"],
    "countrycode": ["country"],
    "displayasentered": [],
    "cellphone": ["cell"],
    "cellphonecountrycode": ["cellcountrycode"],
    "phone": ["home", "homephone"],
    "phonecountrycode": ["phonecountrycode"],
    "email": ["emailaddress"],
    "otheremail": ["email2", "emailaddress2"],
}



import datetime
import gzip
import logging
import time
import uuid
import zipfile
from typing import Optional, Union

import google
import petl
from google.cloud import storage, storage_transfer
from google.oauth2.credentials import Credentials

from parsons.google.utilities import (
    load_google_application_credentials,
    setup_google_application_credentials,
)
from parsons.utilities import files

logger = logging.getLogger(__name__)


class GoogleCloudStorage(object):
    """Google Cloud Storage connector utility

    This class requires application credentials in the form of a
    json or google oauth2 Credentials object. It can be passed in the
    following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.

    * Generate the google credentials object directly, pass in using the
      ``app_creds`` argument.

    For example, to pass in credentials from a parent shell that is
    authenticated with gcloud auth:
    ```
    from google.auth import default

    app_creds, _ = default()

    gcs = GoogleCloudStorage(app_creds=app_creds)
    ```


    `Args:`
        app_creds: str, dict, or google.oauth2.credentials.Credentials object
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set. Can also
            pass a google oauth2 Credentials object directly.
        project: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
    `Returns:`
        GoogleCloudStorage Class

    """

    def __init__(self, app_creds: Optional[Union[str, dict, Credentials]] = None, project=None):
        if isinstance(app_creds, Credentials):
            credentials = app_creds
        else:
            env_credentials_path = str(uuid.uuid4())
            setup_google_application_credentials(
                app_creds, target_env_var_name=env_credentials_path
            )
            credentials = load_google_application_credentials(env_credentials_path)

        self.project = project

        # Throws an error if you pass project=None, so adding if/else statement.
        if not self.project:
            self.client = storage.Client(credentials=credentials)
            """
            Access all methods of `google.cloud` package
            """
        else:
            self.client = storage.Client(credentials=credentials, project=self.project)

    def list_buckets(self):
        """
        Returns a list of buckets

        `Returns:`
            List of buckets
        """

        buckets = [b.name for b in self.client.list_buckets()]
        logger.info(f"Found {len(buckets)}.")
        return buckets

    def bucket_exists(self, bucket_name):
        """
        Verify that a bucket exists

        `Args:`
            bucket_name: str
                The name of the bucket
        `Returns:`
            boolean
        """

        if bucket_name in self.list_buckets():
            logger.debug(f"{bucket_name} exists.")
            return True
        else:
            logger.debug(f"{bucket_name} does not exist.")
            return False

    def get_bucket(self, bucket_name):
        """
        Returns a bucket object

        `Args:`
            bucket_name: str
                The name of bucket
        `Returns:`
            GoogleCloud Storage bucket
        """

        if self.client.lookup_bucket(bucket_name):
            bucket = self.client.get_bucket(bucket_name)
        else:
            raise google.cloud.exceptions.NotFound("Bucket not found")

        logger.debug(f"Returning {bucket_name} object")
        return bucket

    def create_bucket(self, bucket_name):
        """
        Create a bucket.

        `Args:`
            bucket_name: str
                A globally unique name for the bucket.
        `Returns:`
            ``None``
        """

        # TODO: Allow user to set all of the bucket parameters

        self.client.create_bucket(bucket_name)
        logger.info(f"Created {bucket_name} bucket.")

    def delete_bucket(self, bucket_name, delete_blobs=False):
        """
        Delete a bucket. Will fail if not empty unless ``delete_blobs`` argument
        is set to ``True``.

        `Args:`
            bucket_name: str
                The name of the bucket
            delete_blobs: boolean
                Delete blobs in the bucket, if it is not empty
        `Returns:`
            ``None``
        """

        bucket = self.get_bucket(bucket_name)
        bucket.delete(force=delete_blobs)
        logger.info(f"{bucket_name} bucket deleted.")

    def list_blobs(
        self,
        bucket_name,
        max_results=None,
        prefix=None,
        match_glob=None,
        include_file_details=False,
    ):
        """
        List all of the blobs in a bucket

        `Args:`
            bucket_name: str
                The name of the bucket
            max_results: int
                Maximum number of blobs to return
            prefix: str
                A prefix to filter files
            match_glob: str
                Filters files based on glob string. NOTE that the match_glob
                parameter runs on the full blob URI, include a preceding wildcard
                value to account for nested files (*/ for one level, **/ for n levels)
            include_file_details: bool
                If True, returns a list of `Blob` objects with accessible metadata. For
                documentation of attributes associated with `Blob` objects see
                https://cloud.google.com/python/docs/reference/storage/latest/google.cloud.storage.blob.Blob
        `Returns:`
            A list of blob names (or `Blob` objects if `include_file_details` is invoked)
        """

        blobs = self.client.list_blobs(
            bucket_name, max_results=max_results, prefix=prefix, match_glob=match_glob
        )

        if include_file_details:
            lst = list(blobs)
        else:
            lst = [b.name for b in blobs]

        logger.info(f"Found {len(lst)} in {bucket_name} bucket.")

        return lst

    def blob_exists(self, bucket_name, blob_name):
        """
        Verify that a blob exists in the specified bucket

        `Args:`
            bucket_name: str
                The bucket name
            blob_name: str
                The name of the blob
        `Returns:`
            boolean
        """

        if blob_name in self.list_blobs(bucket_name):
            logger.debug(f"{blob_name} exists.")
            return True
        else:
            logger.debug(f"{blob_name} does not exist.")
            return False

    def get_blob(self, bucket_name, blob_name):
        """
        Get a blob object

        `Args`:
            bucket_name: str
                A bucket name
            blob_name: str
                A blob name
        `Returns:`
            A Google Storage blob object
        """

        bucket = self.get_bucket(bucket_name)
        blob = bucket.get_blob(blob_name)
        logger.debug(f"Got {blob_name} object from {bucket_name} bucket.")
        return blob

    def put_blob(self, bucket_name, blob_name, local_path, **kwargs):
        """
        Puts a blob (aka file) in a bucket

        `Args:`
            bucket_name:
                The name of the bucket to store the blob
            blob_name:
                The name of blob to be stored in the bucket
            local_path: str
                The local path of the file to upload
        `Returns:`
            ``None``
        """

        bucket = self.get_bucket(bucket_name)
        blob = storage.Blob(blob_name, bucket)

        with open(local_path, "rb") as f:
            blob.upload_from_file(f, **kwargs)

        logger.info(f"{blob_name} put in {bucket_name} bucket.")

    def download_blob(self, bucket_name, blob_name, local_path=None):
        """
        Gets a blob from a bucket

        `Args:`
            bucket_name: str
                The name of the bucket
            blob_name: str
                The name of the blob
            local_path: str
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.
        `Returns:`
            str
                The path of the downloaded file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path("TEMPTHING")

        bucket = storage.Bucket(self.client, name=bucket_name)
        blob = storage.Blob(blob_name, bucket)

        logger.debug(f"Downloading {blob_name} from {bucket_name} bucket.")
        with open(local_path, "wb") as f:
            blob.download_to_file(f, client=self.client)
        logger.debug(f"{blob_name} saved to {local_path}.")

        return local_path

    def delete_blob(self, bucket_name, blob_name):
        """
        Delete a blob

        `Args:`
            bucket_name: str
                The bucket name
            blob_name: str
                The blob name
        `Returns:`
            ``None``
        """

        blob = self.get_blob(bucket_name, blob_name)
        blob.delete()
        logger.info(f"{blob_name} blob in {bucket_name} bucket deleted.")

    def upload_table(self, table, bucket_name, blob_name, data_type="csv", default_acl=None):
        """
        Load the data from a Parsons table into a blob.

        `Args:`
            table: obj
                A :ref:`parsons-table`
            bucket_name: str
                The name of the bucket to upload the data into.
            blob_name: str
                The name of the blob to upload the data into.
            data_type: str
                The file format to use when writing the data. One of: `csv` or `json`
            default_acl:
                ACL desired for newly uploaded table

        `Returns`:
            String representation of file URI in GCS
        """
        bucket = storage.Bucket(self.client, name=bucket_name)
        blob = storage.Blob(blob_name, bucket)

        if data_type == "csv":
            # If a parsons Table is loaded from a CSV and has had no
            # transformations, the Table.table object will be a petl
            # CSVView. Once any transformations are made, the Table.table
            # becomes a different petl class
            if isinstance(table.table, petl.io.csv_py3.CSVView):
                try:
                    local_file = table.table.source.filename
                except AttributeError:
                    local_file = table.to_csv()
            else:
                local_file = table.to_csv()
            content_type = "text/csv"
        elif data_type == "json":
            local_file = table.to_json()
            content_type = "application/json"
        else:
            raise ValueError(f"Unknown data_type value ({data_type}): must be one of: csv or json")

        try:
            blob.upload_from_filename(
                local_file,
                content_type=content_type,
                client=self.client,
                predefined_acl=default_acl,
            )
        finally:
            files.close_temp_file(local_file)

        return f"gs://{bucket_name}/{blob_name}"

    def get_url(self, bucket_name, blob_name, expires_in=60):
        """
        Generates a presigned url for a blob

        `Args:`
            bucket_name: str
                The name of the bucket
            blob_name: str
                The name of the blob
            expires_in: int
                Minutes until the url expires
        `Returns:`
            url:
                A link to download the object
        """

        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        url = blob.generate_signed_url(
            version="v4",
            expiration=datetime.timedelta(minutes=expires_in),
            method="GET",
        )
        return url

    def copy_bucket_to_gcs(
        self,
        gcs_sink_bucket: str,
        source: str,
        source_bucket: str,
        destination_path: str = "",
        source_path: str = "",
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
    ):
        """
        Creates a one-time transfer job from Amazon S3 to Google Cloud
        Storage. Copies all blobs within the bucket unless a key or prefix
        is passed.

        `Args`:
            gcs_sink_bucket (str):
                Destination for the data transfer (located in GCS)
            source (str):
                File storge vendor [gcs or s3]
            source_bucket (str):
                Source bucket name
            source_path (str):
                Path in the source system pointing to the relevant keys
                / files to sync. Must end in a '/'
            aws_access_key_id (str):
                Access key to authenticate storage transfer
            aws_secret_access_key (str):
                Secret key to authenticate storage transfer
        """
        if source not in ["gcs", "s3"]:
            raise ValueError(f"Blob transfer only supports gcs and s3 sources [source={source}]")
        if source_path and source_path[-1] != "/":
            raise ValueError("Source path much end in a '/'")

        client = storage_transfer.StorageTransferServiceClient()

        now = datetime.datetime.utcnow()
        # Setting the start date and the end date as
        # the same time creates a one-time transfer
        one_time_schedule = {"day": now.day, "month": now.month, "year": now.year}

        if source == "gcs":
            description = f"""One time GCS to GCS Transfer
            [{source_bucket} -> {gcs_sink_bucket}] - {uuid.uuid4()}"""
        elif source == "s3":
            description = f"""One time S3 to GCS Transfer
            [{source_bucket} -> {gcs_sink_bucket}] - {uuid.uuid4()}"""

        transfer_job_config = {
            "project_id": self.project,
            "description": description,
            "status": storage_transfer.TransferJob.Status.ENABLED,
            "schedule": {
                "schedule_start_date": one_time_schedule,
                "schedule_end_date": one_time_schedule,
            },
        }

        # Setup transfer job configuration based on user imput
        if source == "s3":
            blob_storage = "S3"
            transfer_job_config["transfer_spec"] = {
                "aws_s3_data_source": {
                    "bucket_name": source_bucket,
                    "path": source_path,
                    "aws_access_key": {
                        "access_key_id": aws_access_key_id,
                        "secret_access_key": aws_secret_access_key,
                    },
                },
                "gcs_data_sink": {
                    "bucket_name": gcs_sink_bucket,
                    "path": destination_path,
                },
            }
        elif source == "gcs":
            blob_storage = "GCS"
            transfer_job_config["transfer_spec"] = {
                "gcs_data_source": {
                    "bucket_name": source_bucket,
                    "path": source_path,
                },
                "gcs_data_sink": {
                    "bucket_name": gcs_sink_bucket,
                    "path": destination_path,
                },
            }

        create_transfer_job_request = storage_transfer.CreateTransferJobRequest(
            {"transfer_job": transfer_job_config}
        )

        # Create the transfer job
        create_result = client.create_transfer_job(create_transfer_job_request)

        polling = True
        wait_time = 0
        wait_between_attempts_in_sec = 10

        # NOTE: This value defaults to an empty string until GCP
        # triggers the job internally ... we'll use this value to
        # determine whether or not the transfer has kicked off
        latest_operation_name = create_result.latest_operation_name

        while polling:
            if latest_operation_name:
                operation = client.get_operation({"name": latest_operation_name})

                if not operation.done:
                    logger.debug("Operation still running...")

                else:
                    operation_metadata = storage_transfer.TransferOperation.deserialize(
                        operation.metadata.value
                    )
                    error_output = operation_metadata.error_breakdowns
                    if len(error_output) != 0:
                        raise Exception(
                            f"""{blob_storage} to GCS Transfer Job
                            {create_result.name} failed with error: {error_output}"""
                        )
                    else:
                        logger.info(f"TransferJob: {create_result.name} succeeded.")
                        return

            else:
                logger.info("Waiting to kickoff operation...")
                get_transfer_job_request = storage_transfer.GetTransferJobRequest(
                    {"job_name": create_result.name, "project_id": self.project}
                )
                get_result = client.get_transfer_job(request=get_transfer_job_request)
                latest_operation_name = get_result.latest_operation_name

            wait_time += wait_between_attempts_in_sec
            time.sleep(wait_between_attempts_in_sec)

    def format_uri(self, bucket: str, name: str):
        """
        Represent a GCS URI as a string

        `Args`:
            bucket: str
                GCS bucket name
            name: str
                Filename in bucket

        `Returns`:
            String represetnation of URI
        """
        return f"gs://{bucket}/{name}"

    def split_uri(self, gcs_uri: str):
        """
        Split a GCS URI into a bucket and blob name

        `Args`:
            gcs_uri: str
                GCS URI

        `Returns`:
            Tuple of strings with bucket_name and blob_name
        """
        # TODO: make this more robust with regex?
        remove_protocol = gcs_uri.replace("gs://", "")
        uri_parts = remove_protocol.split("/")
        bucket_name = uri_parts[0]
        blob_name = "/".join(uri_parts[1:])
        return bucket_name, blob_name

    def unzip_blob(
        self,
        bucket_name: str,
        blob_name: str,
        compression_type: str = "gzip",
        new_filename: Optional[str] = None,
        new_file_extension: Optional[str] = None,
    ) -> str:
        """
        Downloads and decompresses a blob. The decompressed blob
        is re-uploaded with the same filename if no `new_filename`
        parameter is provided.

        `Args`:
            bucket_name: str
                GCS bucket name

            blob_name: str
                Blob name in GCS bucket

            compression_type: str
                Either `zip` or `gzip`

            new_filename: str
                If provided, replaces the existing blob name
                when the decompressed file is uploaded

            new_file_extension: str
                If provided, replaces the file extension
                when the decompressed file is uploaded

        `Returns`:
            String representation of decompressed GCS URI
        """

        compression_params = {
            "zip": {
                "file_extension": ".zip",
                "compression_function": self.__zip_decompress_and_write_to_gcs,
                "read": "r",
            },
            "gzip": {
                "file_extension": ".gz",
                "compression_function": self.__gzip_decompress_and_write_to_gcs,
            },
        }

        file_extension = compression_params[compression_type]["file_extension"]
        compression_function = compression_params[compression_type]["compression_function"]

        compressed_filepath = self.download_blob(bucket_name=bucket_name, blob_name=blob_name)

        decompressed_filepath = compressed_filepath.replace(file_extension, "")
        decompressed_blob_name = (
            new_filename if new_filename else blob_name.replace(file_extension, "")
        )
        if new_file_extension:
            decompressed_filepath += f".{new_file_extension}"
            decompressed_blob_name += f".{new_file_extension}"

        logger.debug("Decompressing file...")
        compression_function(
            compressed_filepath=compressed_filepath,
            decompressed_filepath=decompressed_filepath,
            decompressed_blob_name=decompressed_blob_name,
            bucket_name=bucket_name,
            new_file_extension=new_file_extension,
        )

        return self.format_uri(bucket=bucket_name, name=decompressed_blob_name)

    def __gzip_decompress_and_write_to_gcs(self, **kwargs):
        """
        Handles `.gzip` decompression and streams blob contents
        to a decompressed storage object
        """

        compressed_filepath = kwargs.pop("compressed_filepath")
        decompressed_blob_name = kwargs.pop("decompressed_blob_name")
        bucket_name = kwargs.pop("bucket_name")

        with gzip.open(compressed_filepath, "rb") as f_in:
            logger.debug(f"Uploading uncompressed file to GCS: {decompressed_blob_name}")
            bucket = self.get_bucket(bucket_name=bucket_name)
            blob = storage.Blob(name=decompressed_blob_name, bucket=bucket)
            blob.upload_from_file(file_obj=f_in, rewind=True, timeout=3600)

    def __zip_decompress_and_write_to_gcs(self, **kwargs):
        """
        Handles `.zip` decompression and streams blob contents
        to a decompressed storage object
        """

        compressed_filepath = kwargs.pop("compressed_filepath")
        decompressed_blob_name = kwargs.pop("decompressed_blob_name")
        decompressed_blob_in_archive = decompressed_blob_name.split("/")[-1]
        bucket_name = kwargs.pop("bucket_name")

        # Unzip the archive
        with zipfile.ZipFile(compressed_filepath) as path_:
            # Open the underlying file
            with path_.open(decompressed_blob_in_archive) as f_in:
                logger.debug(f"Uploading uncompressed file to GCS: {decompressed_blob_name}")
                bucket = self.get_bucket(bucket_name=bucket_name)
                blob = storage.Blob(name=decompressed_blob_name, bucket=bucket)
                blob.upload_from_file(file_obj=f_in, rewind=True, timeout=3600)



import os
import tempfile
import unittest
from abc import ABC
from typing import Optional, Type

from parsons import DBSync, Postgres, Redshift, Table
from parsons.databases.database_connector import DatabaseConnector
from parsons.databases.sqlite import Sqlite
from test.test_databases.fakes import FakeDatabase
from test.utils import assert_matching_tables

_dir = os.path.dirname(__file__)

TEMP_SCHEMA = "parsons_test"


class TestDBSync(ABC, unittest.TestCase):
    setup_sql: Optional[str] = None
    teardown_sql: Optional[str] = None
    temp_schema: Optional[str] = TEMP_SCHEMA
    db: Type[DatabaseConnector]

    @classmethod
    def setUpClass(cls):
        # Skip tests on this abstract base class
        if cls is TestDBSync:
            raise unittest.SkipTest("%s is an abstract base class" % cls.__name__)
        else:
            super(TestDBSync, cls).setUpClass()

    def setUp(self):
        self.initialize_db_connections()

        if self.setup_sql:
            self.source_db.query(self.setup_sql)
            self.destination_db.query(self.setup_sql)

        # Load dummy data to parsons tables
        self.table1 = Table.from_csv(f"{_dir}/test_data/sample_table_1.csv")
        self.table2 = Table.from_csv(f"{_dir}/test_data/sample_table_2.csv")

        self.source_table = (
            f"{self.temp_schema}.source_table" if self.temp_schema else "source_table"
        )
        self.destination_table = (
            f"{self.temp_schema}.destination_table" if self.temp_schema else "destination_table"
        )

        # Create source table
        self.source_db.copy(self.table1, self.source_table, if_exists="truncate")

        self.set_up_db_sync()

    def set_up_db_sync(self, **kwargs) -> None:
        # Create DB Sync object
        self.db_sync = DBSync(self.source_db, self.destination_db, **kwargs)

    def initialize_db_connections(self) -> None:
        self.source_db = self.db()
        self.destination_db = self.db()

    def tearDown(self):
        if self.teardown_sql:
            self.source_db.query(self.teardown_sql)
            self.destination_db.query(self.teardown_sql)

    def assert_matching_tables(self) -> None:
        source = self.source_db.query(f"SELECT * FROM {self.source_table}")
        destination = self.destination_db.query(f"SELECT * FROM {self.destination_table}")
        assert_matching_tables(source, destination)

    def table_sync_full(self, if_exists: str, **kwargs):
        self.db_sync.table_sync_full(
            self.source_table, self.destination_table, if_exists=if_exists, **kwargs
        )

    def test_table_sync_full_drop(self):
        self.table_sync_full(if_exists="drop")
        self.assert_matching_tables()

    def test_table_sync_full_order_by(self):
        self.table_sync_full(if_exists="drop", order_by="data")
        destination_table = self.destination_db.table(self.destination_table)
        rows = destination_table.get_rows()

        # Check that the rows were inserted in the expected order
        self.assertEqual(rows[0]["pk"], "010")
        self.assertEqual(rows[1]["pk"], "012")
        self.assertEqual(rows[2]["pk"], "028")

    def test_table_sync_full_truncate(self):
        self.table_sync_full(if_exists="truncate")
        self.assert_matching_tables()

    def test_table_sync_full_empty_table(self):
        # Empty the source table
        self.source_db.table(self.source_table).truncate()

        # Attempt to sync.
        self.table_sync_full(if_exists="drop", verify_row_count=False)

    def test_table_sync_full_chunk(self):
        # Test chunking in full sync.
        self.db_sync.chunk_size = 10
        self.db_sync.table_sync_full(self.source_table, self.destination_table, if_exists="drop")
        self.assert_matching_tables()

    def test_table_sync_incremental(self):
        # Test that incremental sync

        self.destination_db.copy(self.table1, self.destination_table)
        self.source_db.copy(self.table2, self.source_table, if_exists="append")
        self.db_sync.table_sync_incremental(self.source_table, self.destination_table, "pk")
        self.assert_matching_tables()

    def test_table_sync_incremental_chunk(self):
        # Test chunking of incremental sync.

        self.db_sync.chunk_size = 10
        self.destination_db.copy(self.table1, self.destination_table)
        self.source_db.copy(self.table2, self.source_table, if_exists="append")
        self.db_sync.table_sync_incremental(self.source_table, self.destination_table, "pk")

        self.assert_matching_tables()

    def test_table_sync_incremental_create_destination_table(self):
        # Test that an incremental sync works if the destination table does not exist.
        self.db_sync.table_sync_incremental(self.source_table, self.destination_table, "pk")
        self.assert_matching_tables()

    def test_table_sync_incremental_empty_table(self):
        # Test an incremental sync of a table when the source table is empty.

        # Empty the source table
        self.source_db.table(self.source_table).truncate()

        # Attempt to sync.
        self.db_sync.table_sync_incremental(
            self.source_table, self.destination_table, "pk", verify_row_count=False
        )


class TestFakeDBSync(TestDBSync):
    db = FakeDatabase

    def test_table_sync_full_with_retry(self):
        # Have the copy fail twice
        self.destination_db.setup_table("destination", Table(), failures=2)
        self.set_up_db_sync(retries=2)
        self.table_sync_full(if_exists="drop")
        self.assert_matching_tables()

    def test_table_sync_full_without_retry(self):
        # Have the copy fail
        self.destination_db.setup_table(self.destination_table, Table(), failures=1)
        # Make sure the sync results in an exception
        self.assertRaises(ValueError, lambda: self.table_sync_full(if_exists="drop"))

    def test_table_sync_full_read_chunk(self):
        self.table_sync_full(if_exists="drop")
        self.assert_matching_tables()

        # Make sure copy was called the expected number of times
        # read chunks of 2, 5 rows to write.. should be 3 copy calls
        self.assertEqual(
            len(self.destination_db.copy_call_args[0]),
            3,
            self.destination_db.copy_call_args[0],
        )

    def test_table_sync_full_write_chunk(self):
        self.set_up_db_sync(
            read_chunk_size=1,
            write_chunk_size=3,
        )
        self.table_sync_full(if_exists="drop")
        self.assert_matching_tables()

        # Make sure copy was called the expected number of times
        self.assertEqual(
            len(self.destination_db.copy_call_args[0]),
            3,
            self.destination_db.copy_call_args[0],
        )


class TestSqliteDBSync(TestDBSync):
    db = Sqlite
    temp_schema = None

    def initialize_db_connections(self) -> None:
        self.source_db = self.db(tempfile.mkstemp()[1])
        self.destination_db = self.db(tempfile.mkstemp()[1])


# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestPostgresDBSync(TestDBSync):
    db = Postgres
    setup_sql = f"""
    DROP SCHEMA IF EXISTS {TEMP_SCHEMA} CASCADE;
    CREATE SCHEMA {TEMP_SCHEMA};
    """
    teardown_sql = f"""
    DROP SCHEMA IF EXISTS {TEMP_SCHEMA} CASCADE;
    """


# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestRedshiftDBSync(TestPostgresDBSync):
    db = Redshift



import csv

"""
This utility method is a generalizable method for moving files to an
online file storage class. It is used by methods that require access
to a file via a public url (e.g. VAN). Currently only includes Amazon S3 and
Google Cloud Storage.
"""


def post_file(tbl, type, file_path=None, quoting=csv.QUOTE_MINIMAL, **file_storage_args):
    """
    This utility method is a generalizable method for moving files to an
    online file storage class. It is used by methods that require access
    to a file via a public url (e.g. VAN).

    **S3 is the only option allowed.**

    `Args:`
        tbl: object
            parsons.Table
        type: str
            ``S3`` or ``GCS`` (Google Cloud Storage)
        file_path: str
            The file path to store the file. Not required if provided with
            the **file_storage_args.
        quoting: attr
            The type of quoting to use for the csv.
        **kwargs: kwargs
                Optional arguments specific to the file storage.
    `Returns:`
        ``None``
    """

    if type.upper() == "S3":
        # Overwrite the file_path if key is passed
        if "key" in file_storage_args:
            file_storage_args["key"] = file_path

        return tbl.to_s3_csv(public_url=True, key=file_path, quoting=quoting, **file_storage_args)

    elif type.upper() == "GCS":
        return tbl.to_gcs_csv(
            public_url=True, blob_name=file_path, quoting=quoting, **file_storage_args
        )

    else:
        raise ValueError("Type must be S3 or GCS.")



import os
import unittest

import requests_mock

from parsons import Redash, Table
from parsons.redash.redash import RedashTimeout
from test.utils import assert_matching_tables

BASE_URL = "https://redash.example.com"
API_KEY = "abc123"


class TestRedash(unittest.TestCase):
    mock_data = "foo,bar\n1,2\n3,4"
    mock_data_source = {
        "id": 1,
        "name": "Data Source 1",
        "type": "redshift",
        "options": {
            "dbname": "db_name",
            "host": "host.example.com",
            "password": "--------",
            "port": 5439,
            "user": "username",
        },
    }
    mock_result = Table([("foo", "bar"), ("1", "2"), ("3", "4")])

    def setUp(self):
        self.redash = Redash(BASE_URL, API_KEY)

    @requests_mock.Mocker()
    def test_get_data_source(self, m):
        m.get(f"{BASE_URL}/api/data_sources/1", json=self.mock_data_source)
        assert self.redash.get_data_source(1) == self.mock_data_source

    @requests_mock.Mocker()
    def test_update_data_source(self, m):
        m.post(f"{BASE_URL}/api/data_sources/1", json=self.mock_data_source)
        self.redash.update_data_source(
            1,
            "Data Source 1",
            "redshift",
            "db_name",
            "host.example.com",
            "password",
            5439,
            "username",
        )
        assert m.call_count == 1

    @requests_mock.Mocker()
    def test_cached_query(self, m):
        redash = Redash(BASE_URL)  # no user_api_key
        m.get(f"{BASE_URL}/api/queries/5/results.csv", text=self.mock_data)
        assert_matching_tables(redash.get_cached_query_results(5, API_KEY), self.mock_result)
        self.assertEqual(m._adapter.last_request.path, "/api/queries/5/results.csv")
        self.assertEqual(m._adapter.last_request.query, "api_key=abc123")

        assert_matching_tables(self.redash.get_cached_query_results(5), self.mock_result)
        self.assertEqual(m._adapter.last_request.query, "")

    @requests_mock.Mocker()
    def test_refresh_query(self, m):
        m.post(
            f"{BASE_URL}/api/queries/5/refresh",
            json={"job": {"status": 3, "query_result_id": 21}},
        )
        m.get(f"{BASE_URL}/api/queries/5/results/21.csv", text=self.mock_data)

        assert_matching_tables(
            self.redash.get_fresh_query_results(5, {"yyy": "xxx"}), self.mock_result
        )

    @requests_mock.Mocker()
    def test_refresh_query_poll(self, m):
        m.post(f"{BASE_URL}/api/queries/5/refresh", json={"job": {"id": 66, "status": 1}})
        m.get(
            f"{BASE_URL}/api/jobs/66",
            json={"job": {"id": 66, "status": 3, "query_result_id": 21}},
        )
        m.get(f"{BASE_URL}/api/queries/5/results/21.csv", text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        assert_matching_tables(
            self.redash.get_fresh_query_results(5, {"yyy": "xxx"}), self.mock_result
        )

    @requests_mock.Mocker()
    def test_refresh_query_poll_timeout(self, m):
        m.post(f"{BASE_URL}/api/queries/5/refresh", json={"job": {"id": 66, "status": 1}})
        m.get(f"{BASE_URL}/api/jobs/66", json={"job": {"id": 66, "status": 1}})
        m.get(f"{BASE_URL}/api/queries/5/results/21.csv", text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        self.redash.timeout = 0.01  # timeout
        raised = False
        try:
            self.redash.get_fresh_query_results(5, {"yyy": "xxx"})
        except RedashTimeout:
            raised = True
        self.assertTrue(raised)

    @requests_mock.Mocker()
    def test_to_table(self, m):
        m.post(
            f"{BASE_URL}/api/queries/5/refresh",
            json={"job": {"status": 3, "query_result_id": 21}},
        )
        m.get(f"{BASE_URL}/api/queries/5/results/21.csv", text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        table_data = Redash.load_to_table(
            base_url=BASE_URL,
            user_api_key=API_KEY,
            query_id=5,
            params={"x": "y"},
            verify=False,
        )

        assert_matching_tables(table_data, self.mock_result)

    @requests_mock.Mocker()
    def test_to_table_env_vars(self, m):
        try:
            _environ = dict(os.environ)
            os.environ.update(
                {
                    "REDASH_BASE_URL": BASE_URL,
                    "REDASH_USER_API_KEY": API_KEY,
                    "REDASH_QUERY_ID": "5",
                    "REDASH_QUERY_PARAMS": "p_x=y",
                }
            )
            m.post(
                f"{BASE_URL}/api/queries/5/refresh",
                json={"job": {"status": 3, "query_result_id": 21}},
            )
            m.get(f"{BASE_URL}/api/queries/5/results/21.csv", text=self.mock_data)

            self.redash.pause = 0.01  # shorten pause time

            assert_matching_tables(Redash.load_to_table(), self.mock_result)
        finally:
            os.environ.clear()
            os.environ.update(_environ)



import logging

import braintree

from parsons.etl.table import Table
from parsons.utilities.check_env import check as check_env

logger = logging.getLogger(__name__)


class ParsonsBraintreeError(Exception):
    pass


class Braintree(object):
    """
    Braintree is a payment processor.
    `Args:`
        merchant_id: str
            Braintree merchant id -- probably a 16-char alphanumeric.
            Not required if ``BRAINTREE_MERCHANT_ID`` env variable is set.
        public_key: str
            Braintree public key -- probably a (different) 16-char alphanumeric.
            Not required if ``BRAINTREE_PUBLIC_KEY`` env variable is set.
        private_key: str
            Braintree private key -- probably a 32-char alphanumeric.
            Not required if ``BRAINTREE_PRIVATE_KEY`` env variable is set.
        timeout: int
            Optionally change the timeout from default of 200 seconds.
            Can also be passed with env var ``BRAINTREE_TIMEOUT``.
        production: bool
            Defaults to True.  If you are testing in a Sandbox,
            set this to False.
    `Returns:`
        Braintree class
    """

    query_types = {
        "dispute": braintree.DisputeSearch,
        "transaction": braintree.TransactionSearch,
        "subscription": braintree.SubscriptionSearch,
    }

    credit_card_fields = [
        "bin",
        "card_type",
        "cardholder_name",
        "commercial",
        "country_of_issuance",
        "customer_location",
        "debit",
        "durbin_regulated",
        "expiration_month",
        "expiration_year",
        "healthcare",
        "image_url",
        "issuing_bank",
        "last_4",
        "payroll",
        "prepaid",
        "product_id",
        "token",
        "venmo_sdk",
    ]

    disbursement_fields = [
        "disbursement_date",  # => disbursement_date column
        "funds_held",
        "settlement_amount",
        "settlement_currency_exchange_rate",
        "settlement_currency_iso_code",
        "success",
    ]

    transaction_fields = [
        "additional_processor_response",
        "amount",
        "avs_error_response_code",
        "avs_postal_code_response_code",
        "avs_street_address_response_code",
        "channel",
        "created_at",
        "currency_iso_code",
        "cvv_response_code",
        "discount_amount",
        "escrow_status",
        "gateway_rejection_reason",
        "id",
        "master_merchant_account_id",
        "merchant_account_id",
        "order_id",
        "payment_instrument_type",
        "plan_id",
        "processor_authorization_code",
        "processor_response_code",
        "processor_response_text",
        "processor_settlement_response_code",
        "processor_settlement_response_text",
        "purchase_order_number",
        "recurring",
        "refund_id",
        "refunded_transaction_id",
        "service_fee_amount",
        "settlement_batch_id",
        "shipping_amount",
        "ships_from_postal_code",
        "status",
        "sub_merchant_account_id",
        "subscription_id",
        "tax_amount",
        "tax_exempt",
        "type",
        "updated_at",
        "voice_referral_number",
    ]

    dispute_fields = [
        "id",
        "amount_disputed",
        "amount_won",
        "case_number",
        "currency_iso_code",
        "kind",
        "merchant_account_id",
        "original_dispute_id",
        "processor_comments",
        "reason",
        "reason_code",
        "reason_description",
        "received_date",
        "reference_number",
        "reply_by_date",
        "status",
        # 'transaction.id', # DOT id -- needs to be special-cased (below)
    ]

    subscription_fields = [
        "add_ons",
        "balance",
        "billing_day_of_month",
        "billing_period_end_date",
        "billing_period_start_date",
        "created_at",
        "current_billing_cycle",
        "days_past_due",
        "description",
        # 'descriptor',  # covered under descriptor_fields
        "discounts",
        "failure_count",
        "first_billing_date",
        "id",
        "merchant_account_id",
        "never_expires",
        "next_bill_amount",
        "next_billing_date",
        "next_billing_period_amount",
        "number_of_billing_cycles",
        "paid_through_date",
        "payment_method_token",
        "plan_id",
        "price",
        "status",
        "status_history",
        # 'transactions', # special-cased
        "trial_duration",
        "trial_duration_unit",
        "trial_period",
        "updated_at",
    ]

    descriptor_fields = ["name", "phone", "url"]

    customer_fields = ["first_name", "last_name", "email"]

    def __init__(
        self,
        merchant_id=None,
        public_key=None,
        private_key=None,
        timeout=None,
        production=True,
    ):
        merchant_id = check_env("BRAINTREE_MERCHANT_ID", merchant_id)
        public_key = check_env("BRAINTREE_PUBLIC_KEY", public_key)
        private_key = check_env("BRAINTREE_PRIVATE_KEY", private_key)
        timeout = check_env("BRAINTREE_TIMEOUT", timeout, optional=True) or 200

        self.gateway = braintree.BraintreeGateway(
            braintree.Configuration(
                environment=(
                    braintree.Environment.Production
                    if production
                    else braintree.Environment.Sandbox
                ),
                merchant_id=merchant_id,
                public_key=public_key,
                private_key=private_key,
                timeout=timeout,
            )
        )

    def get_disputes(self, start_date=None, end_date=None, query_list=None, query_dict=None):
        """
        Get a table of disputes based on query parameters.
        There are three ways to pass query arguments: Pass a start_date and end_date
        together for a date range, or pass a query_list or query_dict argument.

        `Args:`
            start_date: date or str
                Start date of the dispute range. Requires `end_date` arg. e.g. '2020-11-03'
            end_date: date or str
                End date of the dispute range. Requires `start_date` arg. e.g. '2020-11-03'
            query_list: list of braintree.DisputeSearch
                You can use the `braintree.DisputeSearch
                <https://developers.braintreepayments.com/reference/request/dispute/search/python>`_
                to create a manual list of query parameters.
            query_dict: jsonable-dict
                query_dict is basically the same as query_list, except instead of using their API
                objects, you can pass it in pure dictionary form.
                Some examples:
                    .. highlight:: python
                    .. code-block:: python

                      # The start_date/end_date arguments are the same as
                      {"effective_date": {"between": [start_date, end_date]}}
                      # some other examples
                      {"merchant_account_id": {"in_list": [123, 456]}}
                      {"created_at": {"greater_than_or_equal": "2020-03-10"}}

        `Returns:`
            Table Class
        """
        collection = self._get_collection(
            "dispute",
            query_list=query_list,
            query_dict=query_dict,
            default_query=(
                {"effective_date": {"between": [start_date, end_date]}}
                if start_date and end_date
                else None
            ),
        )

        # Iterating on collection.items triggers web requests in batches of 50 records
        # Disputes query api doesn't return the ids -- we can't do anything but iterate
        if not collection.is_success:
            raise ParsonsBraintreeError(f"Braintree dispute query failed: {collection.message}")
        return Table(
            [self._dispute_header()] + [self._dispute_to_row(r) for r in collection.disputes.items]
        )

    def get_subscriptions(
        self,
        table_of_ids=None,
        start_date=None,
        end_date=None,
        query_list=None,
        query_dict=None,
        include_transactions=False,
        just_ids=False,
    ):
        """
        Get a table of subscriptions based on query parameters.
        There are three ways to pass query arguments:
        Pass a disbursement_start_date and disbursement_end_date together
        for a date range, or pass a query_list or query_dict argument.

        `Args:`
            start_date: date or str
                Start date of the subscription range. Requires `end_date` arg.
                e.g. '2020-11-03'
            end_date: date or str
                End date of the subscription range. Requires `start_date` arg.
                e.g. '2020-11-03'
            query_list: list of braintree.SubscriptionSearch
                You can use the `braintree.SubscriptionSearch
                <https://developers.braintreepayments.com/reference/request/subscription/search/python>`_
                to create a manual list of query parameters.
            query_dict: jsonable-dict
                query_dict is basically the same as query_list, except instead of using their API
                objects, you can pass it in pure dictionary form.
                Some examples:
                    .. highlight:: python
                    .. code-block:: python

                      # The start_date/end_date arguments are the same as
                      {"created_at": {"between": [start_date, end_date]}}
                      # some other examples
                      {"merchant_account_id": {"in_list": [123, 456]}}
                      {"created_at": {"greater_than_or_equal": "2020-03-10"}}
            include_transactions: bool
                If this is true, include the full collection of transaction objects.
                Otherwise, just return a list of transaction IDs.
            just_ids: bool
                While querying a list of subscription ids is a single, fast query to Braintree's
                API, getting all data for each subscription is force-paginated at 50-records per
                request. If you just need a count or the list of ids, then set `just_ids=True` and
                it will return a single column with `id` instead of all table columns.
            table_of_ids: Table with an `id` column -- i.e. a table returned from `just_ids=True`
                Subsequently, after calling this with `just_ids`, you can prune/alter the ids table
                and then pass the table back to get the full data.
                These are somewhat-niche use-cases, but occasionally crucial
                when a search result returns 1000s of ids.
        `Returns:`
            Table Class
        """
        collection = self._get_collection(
            "subscription",
            table_of_ids=table_of_ids,
            query_list=query_list,
            query_dict=query_dict,
            default_query=(
                {"created_at": {"between": [start_date, end_date]}}
                if start_date and end_date
                else None
            ),
        )
        query_count = len(collection.ids)
        logger.info(
            f"Braintree subscriptions search resulted in subscriptions count of {query_count}"
        )
        if just_ids:
            return Table([("id",)] + [[item_id] for item_id in collection.ids])

        # Iterating on collection.items triggers web requests in batches of 50 records
        # This can be frustratingly slow :-(
        # Also note: Braintree will push you to their new GraphQL API,
        #   but it, too, paginates with a max of 50 records
        logger.debug("Braintree subscriptions iterating to build subscriptions table")
        return Table(
            [self._subscription_header(include_transactions)]
            + [self._subscription_to_row(include_transactions, r) for r in collection.items]
        )

    def get_transactions(
        self,
        table_of_ids=None,
        disbursement_start_date=None,
        disbursement_end_date=None,
        query_list=None,
        query_dict=None,
        just_ids=False,
    ):
        """
        Get a table of transactions based on query parameters.
        There are three ways to pass query arguments:
        Pass a disbursement_start_date and disbursement_end_date together
        for a date range, or pass a query_list or query_dict argument.

        `Args:`
            disbursement_start_date: date or str
                Start date of the disbursement range. Requires `disbursement_end_date` arg.
                e.g. '2020-11-03'
            disbursement_end_date: date or str
                End date of the disbursement range. Requires `disbursement_start_date` arg.
                e.g. '2020-11-03'
            query_list: list of braintree.TransactionSearch
                You can use the `braintree.TransactionSearch
                <https://developers.braintreepayments.com/reference/request/transaction/search/python>`_
                to create a manual list of query parameters.
            query_dict: jsonable-dict
                query_dict is basically the same as query_list, except instead of using their API
                objects, you can pass it in pure dictionary form.
                Some examples:
                    .. highlight:: python
                    .. code-block:: python

                      # The disbursement_start_date/disbursement_end_date arguments are the same as
                      {"disbursement_date": {"between": [start_date, end_date]}}
                      # some other examples
                      {"merchant_account_id": {"in_list": [123, 456]}}
                      {"created_at": {"greater_than_or_equal": "2020-03-10"}}
            just_ids: bool
                While querying a list of transaction ids is a single, fast query to Braintree's API,
                getting all data for each transaction is force-paginated at 50-records per request.
                If you just need a count or the list of ids, then set `just_ids=True` and
                it will return a single column with `id` instead of all table columns.
            table_of_ids: Table with an `id` column -- i.e. a table returned from `just_ids=True`
                Subsequently, after calling this with `just_ids`, you can prune/alter the ids table
                and then pass the table back to get the full data.
                These are somewhat-niche use-cases, but occasionally crucial
                when a search result returns 1000s of ids.
        `Returns:`
            Table Class
        """
        collection = self._get_collection(
            "transaction",
            table_of_ids=table_of_ids,
            query_list=query_list,
            query_dict=query_dict,
            default_query=(
                {"disbursement_date": {"between": [disbursement_start_date, disbursement_end_date]}}
                if disbursement_start_date and disbursement_end_date
                else None
            ),
        )
        query_count = len(collection.ids)
        logger.info(f"Braintree transactions resulted in transaction count of {query_count}")
        if just_ids:
            return Table([("id",)] + [[item_id] for item_id in collection.ids])

        # Iterating on collection.items triggers web requests in batches of 50 records
        # This can be frustratingly slow :-(
        # Also note: Braintree will push you to their new GraphQL API,
        #   but it, too, paginates with a max of 50 records
        logger.debug("Braintree transactions iterating to build transaction table")
        return Table(
            [self._transaction_header()] + [self._transaction_to_row(r) for r in collection.items]
        )

    def _dispute_header(self):
        return self.dispute_fields + ["transaction_id"]

    def _dispute_to_row(self, collection_item):
        row = [getattr(collection_item, k) for k in self.dispute_fields]
        # the single sub-attribute
        row.append(collection_item.transaction.id)
        return row

    def _transaction_header(self):
        return (
            [f"credit_card_{k}" for k in self.credit_card_fields]
            # annoying exception in column name
            + [
                (f"disbursement_{k}" if k != "disbursement_date" else k)
                for k in self.disbursement_fields
            ]
            + [f"customer_{k}" for k in self.customer_fields]
            + self.transaction_fields
        )

    def _transaction_to_row(self, collection_item):
        return (
            [
                (
                    collection_item.credit_card.get(k)
                    if getattr(collection_item, "credit_card", None)
                    else None
                )
                for k in self.credit_card_fields
            ]
            + [getattr(collection_item.disbursement_details, k) for k in self.disbursement_fields]
            + [getattr(collection_item.customer_details, k) for k in self.customer_fields]
            + [getattr(collection_item, k) for k in self.transaction_fields]
        )

    def _subscription_header(self, include_transactions):
        if include_transactions:
            return (
                [f"descriptor_{k}" for k in self.descriptor_fields]
                + self.subscription_fields
                + ["transactions"]
            )
        else:
            return (
                [f"descriptor_{k}" for k in self.descriptor_fields]
                + self.subscription_fields
                + ["transaction_ids"]
            )

    def _subscription_to_row(self, include_transactions, collection_item):
        if include_transactions:
            return (
                [getattr(collection_item.descriptor, k) for k in self.descriptor_fields]
                + [getattr(collection_item, k) for k in self.subscription_fields]
                + [collection_item.transactions]
            )
        else:
            return (
                [getattr(collection_item.descriptor, k) for k in self.descriptor_fields]
                + [getattr(collection_item, k) for k in self.subscription_fields]
                + [";".join(t.id for t in collection_item.transactions)]
            )

    def _get_collection(
        self,
        query_type,
        table_of_ids=None,
        query_list=None,
        query_dict=None,
        default_query=None,
    ):
        collection_query = None
        collection = None
        if query_list:
            collection_query = query_list
        elif query_dict:
            collection_query = self._get_query_objects(query_type, **query_dict)
        elif default_query:
            collection_query = self._get_query_objects(query_type, **default_query)
        if not collection_query:
            raise ParsonsBraintreeError(
                "You must pass some query parameters: "
                "query_dict, start_date with end_date, or query_list"
            )

        if table_of_ids:
            # We don't need to re-do the query, we can just reconstruct the query object
            collection = self._create_collection(
                query_type, table_of_ids.table.values("id"), collection_query
            )
        else:
            collection = getattr(self.gateway, query_type).search(*collection_query)
        return collection

    def _get_query_objects(self, query_type, **queryparams):
        """
        Examples:
        disbursement_date={'between': ['2020-03-20', '2020-03-27']}
        merchant_account_id={'in_list': [123, 456]}
        created_at={'greater_than_or_equal': '2020-03-10'}
        """
        queries = []
        for node, filters in queryparams.items():
            # Very meta programming here, abstracting braintree library's (dumb?) attribute style
            # Example: braintree.DisputeSearch.effective_date
            # Example: braintree.TransactionSearch.disbursement_date
            queryobj = getattr(self.query_types[query_type], node, None)
            if queryobj:
                for qual, vals in filters.items():  # likely only one, but fine
                    queryobj_qualfunc = getattr(queryobj, qual, None)
                    if not queryobj_qualfunc:
                        raise ParsonsBraintreeError("oh no, that's not a braintree parameter")
                    if not isinstance(vals, list):
                        vals = [vals]
                    queries.append(queryobj_qualfunc(*vals))
            else:
                raise ParsonsBraintreeError("oh no, that's not a braintree parameter")
        return queries

    def _create_collection(self, query_type, ids, queries):
        if (query_type == "transaction") or (query_type == "disbursement"):
            gateway = braintree.TransactionGateway(self.gateway)
            return braintree.ResourceCollection(
                queries,
                {"search_results": {"ids": list(ids), "page_size": 50}},
                method=gateway._TransactionGateway__fetch,
            )
        if query_type == "subscription":
            gateway = braintree.SubscriptionGateway(self.gateway)
            return braintree.ResourceCollection(
                queries,
                {"search_results": {"ids": list(ids), "page_size": 50}},
                method=gateway._SubscriptionGateway__fetch,
            )



import json
import logging

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

API_URL = "https://{subdomain}.actionbuilder.org/api/rest/v1"


class ActionBuilder(object):
    """
    `Args:`
        api_token: str
            The OSDI API token
        subdomain: str
            The part of the web app URL preceding '.actionbuilder.org'
        campaign: str
            Optional. The 36-character "interact ID" of the campaign whose data is to be retrieved
            or edited. Can also be supplied in individual methods in case multiple campaigns need
            to be referenced.
    """

    def __init__(self, api_token=None, subdomain=None, campaign=None):
        self.api_token = check_env.check("ACTION_BUILDER_API_TOKEN", api_token)
        self.headers = {
            "Content-Type": "application/json",
            "OSDI-API-Token": self.api_token,
        }
        self.api_url = API_URL.format(subdomain=subdomain)
        self.api = APIConnector(self.api_url, headers=self.headers)
        self.campaign = campaign

    def _campaign_check(self, campaign):
        # Raise an error if campaign is not provided via instatiation nor method argument

        final_campaign = campaign or self.campaign
        if not final_campaign:
            raise ValueError("No campaign provided!")

        return final_campaign

    def _get_page(self, campaign, object_name, page, per_page=25, filter=None):
        # Returns data from one page of results

        campaign = self._campaign_check(campaign)

        if per_page > 25:
            per_page = 25
            logger.info(
                "Action Builder's API will not return more than 25 entries per page. \
            Changing per_page parameter to 25."
            )

        params = {"page": page, "per_page": per_page, "filter": filter}

        url = f"campaigns/{campaign}/{object_name}"

        return self.api.get_request(url=url, params=params)

    def _get_all_records(self, campaign, object_name, limit=None, per_page=25, filter=None):
        # Returns a list of entries for a given object, such as people, tags, or connections.
        # See Action Builder API docs for more: https://www.actionbuilder.org/docs/v1/index.html

        count = 0
        page = 1
        return_list = []

        # Keep getting the next page until record limit is exceeded or an empty result returns
        while True:
            # Get this page and increase page number to the next one
            response = self._get_page(campaign, object_name, page, per_page, filter=filter)
            page = page + 1

            # Check that there's actually data
            response_list = response.get("_embedded", {}).get(f"osdi:{object_name}")

            if not response_list:
                # This page has no data, so we're done
                return Table(return_list)

            # Assuming there's data, add it to the running response list
            return_list.extend(response_list)
            count = count + len(response_list)
            if limit:
                if count >= limit:
                    # Limit reached or exceeded, so return just the requested limit amount
                    return Table(return_list[0:limit])

    def get_campaign_tags(self, campaign=None, limit=None, per_page=25, filter=None):
        """
        Retrieve all tags (i.e. custom field values) within provided limit and filters
        `Args:`
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
            limit: int
                The number of entries to return. When None, returns all entries.
            per_page: int
                The number of entries per page to return. 25 maximum and default.
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            Parsons Table of full set of tags available in Action Builder.
        """

        return self._get_all_records(
            campaign, "tags", limit=limit, per_page=per_page, filter=filter
        )

    def get_tag_by_name(self, tag_name, campaign=None):
        """
        Convenience method to retrieve data on a single tag by its name/value
        `Args:`
            tag_name: str
                The value of the tag to search for.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Parsons Table of data found on tag in Action Builder from searching by name.
        """

        filter = f"name eq '{tag_name}'"

        return self.get_campaign_tags(campaign=campaign, filter=filter)

    def insert_new_tag(self, tag_name, tag_field, tag_section, campaign=None):
        """
        Load a new tag value into Action Builder. Required before applying the value to any entity
        records.
        `Args:`
            tag_name: str
                The name of the new tag, i.e. the custom field value.
            tag_field: str
                The name of the tag category, i.e. the custom field name.
            tag_section: str
                The name of the tag section, i.e. the custom field group name.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict containing Action Builder tag data.
        """

        campaign = self._campaign_check(campaign)
        url = f"campaigns/{campaign}/tags"

        data = {
            "name": tag_name,
            "action_builder:field": tag_field,
            "action_builder:section": tag_section,
        }

        return self.api.post_request(url=url, data=json.dumps(data))

    def _upsert_entity(self, data, campaign):
        # Internal method leveraging the record signup helper endpoint to upsert entity records

        url = f"campaigns/{campaign}/people"

        return self.api.post_request(url=url, data=json.dumps(data))

    def insert_entity_record(self, entity_type, data=None, campaign=None):
        """
        Load a new entity record in Action Builder of the type provided.
        `Args:`
            entity_type: str
                The name of the record type being inserted. Required if identifiers are not
                provided.
            data: dict
                The details to include on the record being upserted, to be included as the value
                of the `person` key. See
                [documentation for the Person Signup Helper](https://www.actionbuilder.org/docs/v1/person_signup_helper.html#post)
                for examples, and
                [the Person endpoint](https://www.actionbuilder.org/docs/v1/people.html#field-names)
                for full entity object composition.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict containing Action Builder entity data.
        """

        name_keys = ("name", "action_builder:name", "given_name")
        error = "Must provide data with name or given_name when inserting new record"
        if not isinstance(data, dict):
            raise ValueError(error)
        name_check = [key for key in data.get("person", {}) if key in name_keys]
        if not name_check:
            raise ValueError(error)

        campaign = self._campaign_check(campaign)

        if not isinstance(data, dict):
            data = {}

        if "person" not in data:
            # The POST data must live inside of person key
            data["person"] = {}

        data["person"]["action_builder:entity_type"] = entity_type

        return self._upsert_entity(data=data, campaign=campaign)

    def update_entity_record(self, identifier, data, campaign=None):
        """
        Update an entity record in Action Builder based on the identifier passed.
        `Args:`
            identifier: str
                The unique identifier for a record being updated. ID strings will need to begin
                with the origin system, followed by a colon, e.g. `action_builder:abc123-...`.
            data: dict
                The details to include on the record being upserted, to be included as the value
                of the `person` key. See
                [documentation for the Person Signup Helper](https://www.actionbuilder.org/docs/v1/person_signup_helper.html#post)
                for examples, and
                [the Person endpoint](https://www.actionbuilder.org/docs/v1/people.html#field-names)
                for full entity object composition.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict containing Action Builder entity data.
        """

        campaign = self._campaign_check(campaign)

        if isinstance(identifier, str):
            # Ensure identifier is a list, even though singular string is called for
            identifier = [identifier]

        # Default to assuming identifier comes from Action Builder and add prefix if missing
        identifiers = [f"action_builder:{id}" if ":" not in id else id for id in identifier]

        if not isinstance(data, dict):
            data = {}

        if "person" not in data:
            # The POST data must live inside of person key
            data["person"] = {}

        data["person"]["identifiers"] = identifiers

        return self._upsert_entity(data=data, campaign=campaign)

    def remove_entity_record_from_campaign(self, identifier, campaign=None):
        """
        Remove an entity record from a campaign. Records cannot be permanently deleted, but a
        record that has been removed from a campaign will not appear in the UI.
        `Args:`
            identifier: str
                The unique identifier for the record being removed. ID strings will need to begin
                with the origin system, followed by a colon, e.g. `action_builder:abc123-...`.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict with HTTP response.
        """

        campaign = self._campaign_check(campaign)

        url = f"campaigns/{campaign}/people/{identifier}"
        return self.api.delete_request(url=url)

    def add_section_field_values_to_record(self, identifier, section, field_values, campaign=None):
        """
        Add one or more tags (i.e. custom field value) to an existing entity record in Action
        Builder. The tags, along with their field and section, must already exist (except for
        date fields).
        `Args:`
            identifier: str
                The unique identifier for a record being updated. ID strings will need to begin
                with the origin system, followed by a colon, e.g. `action_builder:abc123-...`.
            section: str
                The name of the tag section, i.e. the custom field group name.
            field_values: dict
                A collection of field names and tags stored as keys and values.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict containing Action Builder entity data of the entity being tagged.
        """

        tag_data = [
            {
                "action_builder:name": tag,
                "action_builder:field": field,
                "action_builder:section": section,
            }
            for field, tag in field_values.items()
        ]

        data = {"add_tags": tag_data}

        return self.update_entity_record(identifier=identifier, data=data, campaign=campaign)

    def remove_tagging(
        self,
        identifier=None,
        tag_id=None,
        tag_name=None,
        tagging_id=None,
        campaign=None,
    ):
        """
        Remove one or more tags (i.e. custom field value) from an existing entity or connection
        record in Action Builder. The basis for this end point is the combination of the tag's
        interact ID and that of the specific tagging. The tag ID can usually be determined from
        the tag's name, and the tagging ID can be derived if the identifier of the entity or
        connection record is supplied instead.
        `Args:`
            identifier: str
                Optional. The unique identifier for an entity or connection record being updated.
                If omitted, `tagging_id` must be provided.
            tag_id: str
                Optional. The unique identifier for the tag being removed. If omitted, `tag_name`
                must be provided.
            tag_name: str
                Optional. The exact name of the tag being removed. May result in an error if
                multiple tags (in different fields/sections) have the same name. If omitted,
                `tag_id` must be provided.
            tagging_id: str
                Optional. The unique identifier for the specific application of the tag to an
                individual entity or connection record. If omitted, `identifier` must be provided.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            API response JSON which contains `{'message': 'Tag has been removed from Taggable
            Logbook'}` if successful.
        """

        if {tag_name, tag_id} == {None}:
            raise ValueError("Please supply a tag_name or tag_id!")

        if {identifier, tagging_id} == {None}:
            raise ValueError("Please supply an entity or connection identifier, or a tagging id!")

        campaign = self._campaign_check(campaign)
        endpoint = "tags/{}/taggings"

        if tag_name and {tag_id, tagging_id} == {None}:
            tag_data = self.get_tag_by_name(tag_name, campaign=campaign)
            tag_count = tag_data.num_rows

            if tag_count > 1:
                error_msg = f"Found {tag_count} tags with this name. "
                error_msg += "Try with using the unique interact ID"
                raise ValueError(error_msg)

            tag_id = tag_data["identifiers"][0][0].split(":")[1]
            logger.info(f"Tag {tag_name} has ID {tag_id}")

        if tagging_id and not tag_id:
            raise ValueError("Cannot search based on tagging ID alone.")

        if tag_id and not tagging_id:
            taggings = self._get_all_records(self.campaign, endpoint.format(tag_id))
            taggings_filtered = taggings.select_rows(
                lambda row: (
                    identifier in row["_links"]["action_builder:connection"]["href"]
                    if row["item_type"] == "connection"
                    else identifier in row["osdi:person"]["href"]
                )
            )
            tagging_id = [
                x.split(":")[1]
                for x in taggings_filtered["identifiers"][0]
                if "action_builder" in x
            ][0]

        logger.info(f"Removing tag {tag_id} from {identifier or tagging_id}")
        return self.api.delete_request(
            f"campaigns/{campaign}/{endpoint.format(tag_id)}/{tagging_id}"
        )

    def upsert_connection(self, identifiers, tag_data=None, campaign=None, reactivate=True):
        """
        Load or update a connection record in Action Builder between two existing entity records.
        Only one connection record is allowed per pair of entities, so if the connection already
        exists, this method will update, but will otherwise create a new connection record.
        `Args:`
            identifiers: list
                A list of two unique identifier strings for records being connected. ID strings
                will need to begin with the origin system, followed by a colon, e.g.
                `action_builder:abc123-...`. Requires exactly two identifiers.
            tag_data: list
                List of dicts of tags to be added to the connection record (i.e. Connection Info).
                See [documentation on Connection Helper](https://www.actionbuilder.org/docs/v1/connection_helper.html#post)
                for examples.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
            reactivate: bool
                Optional. Whether or not to set the `inactive` flag on a given Connection to False
                if the Connection exists and has `inactive` set to True. True by default.
        `Returns:`
            Dict containing Action Builder connection data.
        """

        # Check that there are exactly two identifiers and that campaign is provided first
        if not isinstance(identifiers, list):
            raise ValueError("Must provide identifiers as a list")

        if len(identifiers) != 2:
            raise ValueError("Must provide exactly two identifiers")

        campaign = self._campaign_check(campaign)

        url = f"campaigns/{campaign}/people/{identifiers[0]}/connections"

        data = {
            "connection": {
                # person_id is used even if entity is not Person
                "person_id": identifiers[1]
            }
        }

        if reactivate:
            data["connection"]["inactive"] = False

        if tag_data:
            if isinstance(tag_data, dict):
                tag_data = [tag_data]

            if not isinstance(tag_data[0], dict):
                raise ValueError("Must provide tag_data as a dict or list of dicts")

            data["add_tags"] = tag_data

        return self.api.post_request(url=url, data=json.dumps(data))

    def deactivate_connection(
        self,
        from_identifier,
        connection_identifier=None,
        to_identifier=None,
        campaign=None,
    ):
        """
        Deactivate an existing connection record in Action Builder between two existing entity
        records. Only one connection record is allowed per pair of entities, so this can be done
        by supplying the ID for the connection record, or for the two connected entity records.
        `Args:`
            from_identifier: str
                Unique identifier for one of the two entities with a connection.
            connection_identifier: str
                Optional. The unique identifier for an entity or connection record being updated.
                If omitted, `to_identifier` must be provided.
            to_identifier: str
                Optional. The second entity with a connection to `from_entity`. If omitted,
                `connection_identifier` must be provided.
            campaign: str
                Optional. The 36-character "interact ID" of the campaign whose data is to be
                retrieved or edited. Not necessary if supplied when instantiating the class.
        `Returns:`
            Dict containing Action Builder connection data.
        """

        # Check that either connection or second entity identifier are provided
        if {connection_identifier, to_identifier} == {None}:
            raise ValueError("Must provide a connection ID or an ID for the second entity")

        campaign = self._campaign_check(campaign)

        url = f"campaigns/{campaign}/people/{from_identifier}/connections"

        data = {"connection": {"inactive": True}}

        # Prioritize connection ID to avoid potential confusion if to_identifier is also provided
        # to_identifier entity could have duplicates, connection ID is more specific
        if connection_identifier:
            url += f"/{connection_identifier}"

            # Despite the documentation, PUT requests don't require the outer "connection" key
            data = data["connection"]

            return self.api.put_request(url=url, data=json.dumps(data))

        # If no connection ID then there must be a to_identifier not to have errored by now
        else:
            data["connection"]["person_id"] = to_identifier
            return self.api.post_request(url=url, data=json.dumps(data))



import datetime
import json
import logging
import pickle
import random
import uuid
from contextlib import contextmanager
from typing import List, Optional, Union

import google
import petl
from google.api_core import exceptions
from google.cloud import bigquery
from google.cloud.bigquery import dbapi, job
from google.cloud.bigquery.job import ExtractJobConfig, LoadJobConfig
from google.oauth2.credentials import Credentials

from parsons.databases.database_connector import DatabaseConnector
from parsons.databases.table import BaseTable
from parsons.etl import Table
from parsons.google.google_cloud_storage import GoogleCloudStorage
from parsons.google.utilities import (
    load_google_application_credentials,
    setup_google_application_credentials,
)
from parsons.utilities import check_env
from parsons.utilities.files import create_temp_file

logger = logging.getLogger(__name__)

BIGQUERY_TYPE_MAP = {
    "str": "STRING",
    "float": "FLOAT",
    "int": "INTEGER",
    "bool": "BOOLEAN",
    "datetime": "DATETIME",
    "date": "DATE",
    "time": "TIME",
    "NoneType": "STRING",
    "UUID": "STRING",
    "timestamp": "TIMESTAMP",
    "Decimal": "FLOAT",
}

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000


def parse_table_name(table_name):
    # Helper function to parse out the different components of a table ID
    parts = table_name.split(".")
    parts.reverse()
    parsed = {
        "project": None,
        "dataset": None,
        "table": None,
    }
    if len(parts) > 0:
        parsed["table"] = parts[0]
    if len(parts) > 1:
        parsed["dataset"] = parts[1]
    if len(parts) > 2:
        parsed["project"] = parts[2]
    return parsed


def ends_with_semicolon(query: str) -> str:
    query = query.strip()
    if query[-1] == ";":
        return query
    return query + ";"


def map_column_headers_to_schema_field(schema_definition: list) -> list:
    """
    Loops through a list of dictionaries and instantiates
    google.cloud.bigquery.SchemaField objects. Useful docs
    from Google's API can be found here:
        https://cloud.google.com/python/docs/reference/bigquery/latest/google.cloud.bigquery.schema.SchemaField

    `Args`:
        schema_definition: list
        This function expects a list of dictionaries in the following format:

        ```
        schema_definition = [
            {
                "name": column_name,
                "field_type": [INTEGER, STRING, FLOAT, etc.]
            },
            {
                "name": column_name,
                "field_type": [INTEGER, STRING, FLOAT, etc.],
                "mode": "REQUIRED"
            },
            {
                "name": column_name,
                "field_type": [INTEGER, STRING, FLOAT, etc.],
                "default_value_expression": CURRENT_TIMESTAMP()
            }
        ]
        ```

    `Returns`:
        List of instantiated `SchemaField` objects
    """

    # TODO - Better way to test for this
    if isinstance(schema_definition[0], bigquery.SchemaField):
        logger.debug("User supplied list of SchemaField objects")
        return schema_definition

    return [bigquery.SchemaField(**x) for x in schema_definition]


class GoogleBigQuery(DatabaseConnector):
    """
    Class for querying BigQuery table and returning the data as Parsons tables.

    This class requires application credentials in the form of a json. It can be passed
    in the following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.

    Args:
        app_creds: str
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        project: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
        location: str
            Default geographic location for tables
        client_options: dict
            A dictionary containing any requested client options. Defaults to the required
            scopes for making API calls against External tables stored in Google Drive.
            Can be set to None if these permissions are not desired
        gcs_temp_bucket: str
            Name of the GCS bucket that will be used for storing data during bulk transfers.
            Required if you intend to perform bulk data transfers (eg. the copy_from_gcs method),
            and env variable ``GCS_TEMP_BUCKET`` is not populated.
    """

    def __init__(
        self,
        app_creds: Optional[Union[str, dict, Credentials]] = None,
        project=None,
        location=None,
        client_options: dict = None,
        tmp_gcs_bucket: Optional[str] = None,
    ):
        if client_options is None:
            client_options = {
                "scopes": [
                    "https://www.googleapis.com/auth/drive",
                    "https://www.googleapis.com/auth/bigquery",
                    "https://www.googleapis.com/auth/cloud-platform",
                ]
            }
        self.app_creds = app_creds

        if isinstance(app_creds, Credentials):
            self.credentials = app_creds
        else:
            self.env_credential_path = str(uuid.uuid4())
            setup_google_application_credentials(
                app_creds, target_env_var_name=self.env_credential_path
            )
            self.credentials = load_google_application_credentials(self.env_credential_path)

        self.project = project
        self.location = location
        self.client_options = client_options
        self.tmp_gcs_bucket = tmp_gcs_bucket

        # We will not create the client until we need to use it, since creating the client
        # without valid GOOGLE_APPLICATION_CREDENTIALS raises an exception.
        # This attribute will be used to hold the client once we have created it.
        self._client = None

        self._dbapi = dbapi

        self.dialect = "bigquery"

    @property
    def client(self):
        """
        Get the Google BigQuery client to use for making queries.

        `Returns:`
            `google.cloud.bigquery.client.Client`
        """
        if not self._client:
            # Create a BigQuery client to use to make the query
            self._client = bigquery.Client(
                project=self.project,
                location=self.location,
                client_options=self.client_options,
                credentials=self.credentials,
            )

        return self._client

    @contextmanager
    def connection(self):
        """
        Generate a BigQuery connection.
        The connection is set up as a python "context manager", so it will be closed
        automatically when the connection goes out of scope. Note that the BigQuery
        API uses jobs to run database operations and, as such, simply has a no-op for
        a "commit" function.

        If you would like to manage transactions, please use multi-statement queries
        as [outlined here](https://cloud.google.com/bigquery/docs/transactions)
        or utilize the `query_with_transaction` method on this class.

        When using the connection, make sure to put it in a ``with`` block (necessary for
        any context manager):
        ``with bq.connection() as conn:``

        `Returns:`
            Google BigQuery ``connection`` object
        """
        conn = self._dbapi.connect(self.client)
        try:
            yield conn
        finally:
            conn.close()

    @contextmanager
    def cursor(self, connection):
        cur = connection.cursor()
        try:
            yield cur
        finally:
            cur.close()

    def query(
        self,
        sql: str,
        parameters: Optional[Union[list, dict]] = None,
        return_values: bool = True,
    ) -> Optional[Table]:
        """
        Run a BigQuery query and return the results as a Parsons table.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the BigQuery style where parameters are prefixed with `@`s.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

           name = "Beatrice O'Brady"
           sql = 'SELECT * FROM my_table WHERE name = %s'
           rs.query(sql, parameters=[name])

        .. code-block:: python

           name = "Beatrice O'Brady"
           sql = "SELECT * FROM my_table WHERE name = %(name)s"
           rs.query(sql, parameters={'name': name})

        `Args:`
            sql: str
                A valid BigTable statement
            parameters: dict
                A dictionary of query parameters for BigQuery.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        with self.connection() as connection:
            return self.query_with_connection(
                sql, connection, parameters=parameters, return_values=return_values
            )

    def query_with_connection(
        self, sql, connection, parameters=None, commit=True, return_values: bool = True
    ):
        """
        Execute a query against the BigQuery database, with an existing connection.
        Useful for batching queries together. Will return ``None`` if the query
        returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``redshift.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Must be true. BigQuery

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if not commit:
            raise ValueError(
                """
                BigQuery implementation uses an API client which always auto-commits.
                If you wish to wrap multiple queries in a transaction, use
                Mulit-Statement transactions within a single query as outlined
                here: https://cloud.google.com/bigquery/docs/transactions or use the
                `query_with_transaction` method on the BigQuery connector.
            """
            )

        # get our connection and cursor
        with self.cursor(connection) as cursor:
            # Run the query
            cursor.execute(sql, parameters)

            if not return_values:
                return None

            # This applies when running a SQL statement without any return value
            # e.g. when creating a view or a table
            # This does not apply when 0 rows are returned
            if not cursor.description:
                return None

            final_table = self._fetch_query_results(cursor=cursor)

            return final_table

    def query_with_transaction(self, queries, parameters=None):
        queries_with_semicolons = [ends_with_semicolon(q) for q in queries]
        queries_on_newlines = "\n".join(queries_with_semicolons)
        queries_wrapped = f"""
        BEGIN
            BEGIN TRANSACTION;
            {queries_on_newlines}
            COMMIT TRANSACTION;
        END;
        """
        self.query(sql=queries_wrapped, parameters=parameters, return_values=False)

    def get_job(
        self, job_id: str, **job_kwargs
    ) -> Union[job.LoadJob, job.CopyJob, job.ExtractJob, job.QueryJob, job.UnknownJob]:
        """
        Fetch a job

        `Args:`
            job_id: str
                ID of job to fetch
            location: str
                Location where the job was run
            **job_kwargs: kwargs
                Other arguments to pass to the underlying get_job
                call on the BigQuery client.
        """
        return self.client.get_job(job_id=job_id, **job_kwargs)

    def copy_from_gcs(
        self,
        gcs_blob_uri: str,
        table_name: str,
        if_exists: str = "fail",
        max_errors: int = 0,
        data_type: str = "csv",
        csv_delimiter: str = ",",
        ignoreheader: int = 1,
        nullas: Optional[str] = None,
        allow_quoted_newlines: bool = True,
        allow_jagged_rows: bool = True,
        quote: Optional[str] = None,
        schema: Optional[List[dict]] = None,
        job_config: Optional[LoadJobConfig] = None,
        force_unzip_blobs: bool = False,
        compression_type: str = "gzip",
        new_file_extension: str = "csv",
        template_table: Optional[str] = None,
        max_timeout: int = 21600,
        **load_kwargs,
    ):
        """
        Copy a csv saved in Google Cloud Storage into Google BigQuery.

        `Args:`
            gcs_blob_uri: str
                The GoogleCloudStorage URI referencing the file to be copied.
            table_name: str
                The table name to load the data into. Will be used to generate load schema
                if no custom schema or template table are supplied and the if_exists is
                set to "truncate" or "append".
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table. This maps to `write_disposition` in the
                `LoadJobConfig` class.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails. This maps to `max_bad_records` in the `LoadJobConfig` class.
            data_type: str
                Denotes whether target file is a JSON or CSV
            csv_delimiter: str
                Character used to separate values in the target file
            ignoreheader: int
                Treats the specified number_rows as a file header and doesn't load them
            nullas: str
                Loads fields that match null_string as NULL, where null_string can be any string
            allow_quoted_newlines: bool
                If True, detects quoted new line characters within a CSV field and does
                not interpret the quoted new line character as a row boundary
            allow_jagged_rows: bool
                Allow missing trailing optional columns (CSV only).
            quote: str
                The value that is used to quote data sections in a CSV file.
                BigQuery converts the string to ISO-8859-1 encoding, and then uses the first byte of
                the encoded string to split the data in its raw, binary state.
            schema: list
                BigQuery expects a list of dictionaries in the following format
                ```
                schema = [
                    {"name": "column_name", "type": STRING},
                    {"name": "another_column_name", "type": INT}
                ]
                ```
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided. Note
                if there are any conflicts between the job_config and other parameters, the
                job_config values are preferred.
            force_unzip_blobs: bool
                If True, target blobs will be unzipped before being loaded to BigQuery.
            compression_type: str
                Accepts `zip` or `gzip` values to differentially unzip a compressed
                blob in cloud storage.
            new_file_extension: str
                Provides a file extension if a blob is decompressed and rewritten
                to cloud storage.
            template_table: str
                Table name to be used as the load schema. Load operation wil use the same
                columns and data types as the template table.
            max_timeout: int
                The maximum number of seconds to wait for a request before the job fails.
            **load_kwargs: kwargs
                Other arguments to pass to the underlying load_table_from_uri
                call on the BigQuery client.
        """
        self._validate_copy_inputs(if_exists=if_exists, data_type=data_type)

        job_config = self._process_job_config(
            job_config=job_config,
            destination_table_name=table_name,
            if_exists=if_exists,
            max_errors=max_errors,
            data_type=data_type,
            csv_delimiter=csv_delimiter,
            ignoreheader=ignoreheader,
            nullas=nullas,
            allow_quoted_newlines=allow_quoted_newlines,
            allow_jagged_rows=allow_jagged_rows,
            quote=quote,
            custom_schema=schema,
            template_table=template_table,
        )

        # load CSV from Cloud Storage into BigQuery
        table_ref = self.get_table_ref(table_name=table_name)

        try:
            if force_unzip_blobs:
                return self.copy_large_compressed_file_from_gcs(
                    gcs_blob_uri=gcs_blob_uri,
                    table_name=table_name,
                    if_exists=if_exists,
                    max_errors=max_errors,
                    data_type=data_type,
                    csv_delimiter=csv_delimiter,
                    ignoreheader=ignoreheader,
                    nullas=nullas,
                    allow_quoted_newlines=allow_quoted_newlines,
                    quote=quote,
                    schema=schema,
                    job_config=job_config,
                    compression_type=compression_type,
                    new_file_extension=new_file_extension,
                    max_timeout=max_timeout,
                )
            else:
                return self._load_table_from_uri(
                    source_uris=gcs_blob_uri,
                    destination=table_ref,
                    job_config=job_config,
                    max_timeout=max_timeout,
                    **load_kwargs,
                )
        except exceptions.BadRequest as e:
            if "one of the files is larger than the maximum allowed size." in str(e):
                logger.debug(
                    f"{gcs_blob_uri.split('/')[-1]} exceeds max size ... \
                    running decompression function..."
                )

                return self.copy_large_compressed_file_from_gcs(
                    gcs_blob_uri=gcs_blob_uri,
                    table_name=table_name,
                    if_exists=if_exists,
                    max_errors=max_errors,
                    data_type=data_type,
                    csv_delimiter=csv_delimiter,
                    ignoreheader=ignoreheader,
                    nullas=nullas,
                    allow_quoted_newlines=allow_quoted_newlines,
                    quote=quote,
                    schema=schema,
                    job_config=job_config,
                    compression_type=compression_type,
                    new_file_extension=new_file_extension,
                    max_timeout=max_timeout,
                )
            elif "Schema has no field" in str(e):
                logger.debug(f"{gcs_blob_uri.split('/')[-1]} is empty, skipping file")
                return "Empty file"

            else:
                raise e

        except exceptions.DeadlineExceeded as e:
            logger.error(f"Max timeout exceeded for {gcs_blob_uri.split('/')[-1]}")
            raise e

    def copy_large_compressed_file_from_gcs(
        self,
        gcs_blob_uri: str,
        table_name: str,
        if_exists: str = "fail",
        max_errors: int = 0,
        data_type: str = "csv",
        csv_delimiter: str = ",",
        ignoreheader: int = 1,
        nullas: Optional[str] = None,
        allow_quoted_newlines: bool = True,
        allow_jagged_rows: bool = True,
        quote: Optional[str] = None,
        schema: Optional[List[dict]] = None,
        job_config: Optional[LoadJobConfig] = None,
        compression_type: str = "gzip",
        new_file_extension: str = "csv",
        template_table: Optional[str] = None,
        max_timeout: int = 21600,
        **load_kwargs,
    ):
        """
        Copy a compressed CSV file that exceeds the maximum size in Google Cloud Storage
        into Google BigQuery.

        `Args:`
            gcs_blob_uri: str
                The GoogleCloudStorage URI referencing the file to be copied.
            table_name: str
                The table name to load the data into. Will be used to generate load schema
                if no custom schema or template table are supplied and the if_exists is
                set to "truncate" or "append".
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table. This maps to `write_disposition` in the
                `LoadJobConfig` class.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails. This maps to `max_bad_records` in the `LoadJobConfig` class.
            data_type: str
                Denotes whether target file is a JSON or CSV
            csv_delimiter: str
                Character used to separate values in the target file
            ignoreheader: int
                Treats the specified number_rows as a file header and doesn't load them
            nullas: str
                Loads fields that match null_string as NULL, where null_string can be any string
            allow_quoted_newlines: bool
                If True, detects quoted new line characters within a CSV field
                and does not interpret the quoted new line character as a row boundary
            allow_jagged_rows: bool
                Allow missing trailing optional columns (CSV only).
            quote: str
                The value that is used to quote data sections in a CSV file.
                BigQuery converts the string to ISO-8859-1 encoding, and then uses the first byte of
                the encoded string to split the data in its raw, binary state.
            schema: list
                BigQuery expects a list of dictionaries in the following format
                ```
                schema = [
                    {"name": "column_name", "type": STRING},
                    {"name": "another_column_name", "type": INT}
                ]
                ```
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided. Note
                if there are any conflicts between the job_config and other parameters, the
                job_config values are preferred.
            compression_type: str
                Accepts `zip` or `gzip` values to differentially unzip a compressed
                blob in cloud storage.
            new_file_extension: str
                Provides a file extension if a blob is decompressed and rewritten to cloud storage.
            template_table: str
                Table name to be used as the load schema. Load operation wil use the same
                columns and data types as the template table.
            max_timeout: int
                The maximum number of seconds to wait for a request before the job fails.
            **load_kwargs: kwargs
                Other arguments to pass to the underlying load_table_from_uri call on the BigQuery
                client.
        """

        self._validate_copy_inputs(if_exists=if_exists, data_type=data_type)

        job_config = self._process_job_config(
            job_config=job_config,
            destination_table_name=table_name,
            if_exists=if_exists,
            max_errors=max_errors,
            data_type=data_type,
            csv_delimiter=csv_delimiter,
            ignoreheader=ignoreheader,
            nullas=nullas,
            allow_quoted_newlines=allow_quoted_newlines,
            allow_jagged_rows=allow_jagged_rows,
            quote=quote,
            custom_schema=schema,
            template_table=template_table,
        )

        # TODO - See if this inheritance is happening in other places
        gcs = GoogleCloudStorage(app_creds=self.app_creds, project=self.project)
        old_bucket_name, old_blob_name = gcs.split_uri(gcs_uri=gcs_blob_uri)

        uncompressed_gcs_uri = None

        try:
            logger.debug("Unzipping large file")
            uncompressed_gcs_uri = gcs.unzip_blob(
                bucket_name=old_bucket_name,
                blob_name=old_blob_name,
                new_file_extension=new_file_extension,
                compression_type=compression_type,
            )

            logger.debug(f"Loading uncompressed uri into BigQuery {uncompressed_gcs_uri}...")
            table_ref = self.get_table_ref(table_name=table_name)
            return self._load_table_from_uri(
                source_uris=uncompressed_gcs_uri,
                destination=table_ref,
                job_config=job_config,
                max_timeout=max_timeout,
                **load_kwargs,
            )

        finally:
            if uncompressed_gcs_uri:
                new_bucket_name, new_blob_name = gcs.split_uri(gcs_uri=uncompressed_gcs_uri)
                gcs.delete_blob(new_bucket_name, new_blob_name)
                logger.debug("Successfully dropped uncompressed blob")

    def copy_s3(
        self,
        table_name,
        bucket,
        key,
        if_exists: str = "fail",
        max_errors: int = 0,
        data_type: str = "csv",
        csv_delimiter: str = ",",
        ignoreheader: int = 1,
        nullas: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        gcs_client: Optional[GoogleCloudStorage] = None,
        tmp_gcs_bucket: Optional[str] = None,
        template_table: Optional[str] = None,
        job_config: Optional[LoadJobConfig] = None,
        max_timeout: int = 21600,
        **load_kwargs,
    ):
        """
        Copy a file from s3 to BigQuery.

        `Args:`
            table_name: str
                The table name and schema (``tmc.cool_table``) to point the file.
            bucket: str
                The s3 bucket where the file or manifest is located.
            key: str
                The key of the file or manifest in the s3 bucket.
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails.
            data_type: str
                The data type of the file. Only ``csv`` supported currently.
            csv_delimiter: str
                The delimiter of the ``csv``. Only relevant if data_type is ``csv``.
            ignoreheader: int
                The number of header rows to skip. Ignored if data_type is ``json``.
            nullas: str
                Loads fields that match string as NULL
            aws_access_key_id:
                An AWS access key granted to the bucket where the file is located. Not required
                if keys are stored as environmental variables.
            aws_secret_access_key:
                An AWS secret access key granted to the bucket where the file is located. Not
                required if keys are stored as environmental variables.
            gcs_client: object
                The GoogleCloudStorage Connector to use for loading data into Google Cloud Storage.
            tmp_gcs_bucket: str
                The name of the Google Cloud Storage bucket to use to stage the data to load
                into BigQuery. Required if `GCS_TEMP_BUCKET` is not specified or set on
                the class instance.
            template_table: str
                Table name to be used as the load schema. Load operation wil use the same
                columns and data types as the template table.
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided. Note
                if there are any conflicts between the job_config and other parameters, the
                job_config values are preferred.
            max_timeout: int
                The maximum number of seconds to wait for a request before the job fails.

        `Returns`
            Parsons Table or ``None``
                See :ref:`parsons-table` for output options.
        """

        # copy from S3 to GCS
        tmp_gcs_bucket = (
            tmp_gcs_bucket
            or self.tmp_gcs_bucket
            or check_env.check("GCS_TEMP_BUCKET", tmp_gcs_bucket)
        )
        gcs_client = gcs_client or GoogleCloudStorage()
        temp_blob_uri = gcs_client.copy_s3_to_gcs(
            aws_source_bucket=bucket,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            gcs_sink_bucket=tmp_gcs_bucket,
            aws_s3_key=key,
        )
        temp_blob_name = key
        temp_blob_uri = gcs_client.format_uri(bucket=tmp_gcs_bucket, name=temp_blob_name)

        # load CSV from Cloud Storage into BigQuery
        try:
            return self.copy_from_gcs(
                gcs_blob_uri=temp_blob_uri,
                table_name=table_name,
                if_exists=if_exists,
                max_errors=max_errors,
                data_type=data_type,
                csv_delimiter=csv_delimiter,
                ignoreheader=ignoreheader,
                nullas=nullas,
                job_config=job_config,
                template_table=template_table,
                max_timeout=max_timeout,
                **load_kwargs,
            )
        finally:
            gcs_client.delete_blob(tmp_gcs_bucket, temp_blob_name)

    def copy_direct(
        self,
        tbl: Table,
        table_name: str,
        if_exists: str = "fail",
        max_errors: int = 0,
        job_config: Optional[LoadJobConfig] = None,
        template_table: Optional[str] = None,
        ignoreheader: int = 1,
        nullas: Optional[str] = None,
        allow_quoted_newlines: bool = True,
        allow_jagged_rows: bool = True,
        quote: Optional[str] = None,
        schema: Optional[List[dict]] = None,
        max_timeout: int = 21600,
        convert_dict_list_columns_to_json: bool = True,
        **load_kwargs,
    ):
        """
        Copy a :ref:`parsons-table` into Google BigQuery
        directly. This will work well for smaller data. For larger
        data, use the :meth:`copy` method which stages the upload through CloudStorage.

        `Args:`
            tbl: obj
                The Parsons Table to copy into BigQuery.
            table_name: str
                The table name to load the data into. Will be used to generate load schema
                if no custom schema or template table are supplied and if_exists is
                set to "truncate" or "append".
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails.
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided.
            template_table: str
                Table name to be used as the load schema. Load operation wil use the same
                columns and data types as the template table.
            max_timeout: int
                The maximum number of seconds to wait for a request before the job fails.
            convert_dict_list_columns_to_json: bool
                If set to True, will convert any dict or list columns (which cannot by default be successfully loaded to BigQuery to JSON strings)
            **load_kwargs: kwargs
                Arguments to pass to the underlying load_table_from_uri call on the BigQuery
                client.

        """

        if convert_dict_list_columns_to_json:
            tbl = self._stringify_records(tbl)

        job_config = self._prepare_local_upload_job(
            tbl,
            table_name,
            if_exists,
            max_errors,
            job_config,
            template_table,
            ignoreheader,
            nullas,
            allow_quoted_newlines,
            allow_jagged_rows,
            quote,
            schema,
        )

        tmpfile_path = tbl.to_csv()
        with open(tmpfile_path, mode="rb") as tmpfile:
            load_job = self.client.load_table_from_file(
                tmpfile,
                destination=self.get_table_ref(table_name=table_name),
                job_config=job_config,
            )

            try:
                load_job.result(timeout=max_timeout)
                return load_job
            except exceptions.BadRequest as e:
                for idx, error_ in enumerate(load_job.errors):
                    if idx == 0:
                        logger.error("* Load job failed. Enumerating errors collection below:")
                    logger.error(f"** Error collection - index {idx}:")
                    logger.error(error_)

                raise e

    def copy(
        self,
        tbl: Table,
        table_name: str,
        if_exists: str = "fail",
        max_errors: int = 0,
        tmp_gcs_bucket: Optional[str] = None,
        gcs_client: Optional[GoogleCloudStorage] = None,
        job_config: Optional[LoadJobConfig] = None,
        template_table: Optional[str] = None,
        ignoreheader: int = 1,
        nullas: Optional[str] = None,
        allow_quoted_newlines: bool = True,
        allow_jagged_rows: bool = True,
        quote: Optional[str] = None,
        schema: Optional[List[dict]] = None,
        max_timeout: int = 21600,
        convert_dict_list_columns_to_json: bool = True,
        **load_kwargs,
    ):
        """
        Copy a :ref:`parsons-table` into Google BigQuery via Google Cloud Storage.

        `Args:`
            tbl: obj
                The Parsons Table to copy into BigQuery.
            table_name: str
                The table name to load the data into. Will be used to generate load schema
                if no custom schema or template table are supplied and if_exists is
                set to "truncate" or "append".
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails.
            tmp_gcs_bucket: str
                The name of the Google Cloud Storage bucket to use to stage the data to load
                into BigQuery. Required if `GCS_TEMP_BUCKET` is not specified or set on
                the class instance.
            gcs_client: object
                The GoogleCloudStorage Connector to use for loading data into Google Cloud Storage.
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided.
            template_table: str
                Table name to be used as the load schema. Load operation wil use the same
                columns and data types as the template table.
            max_timeout: int
                The maximum number of seconds to wait for a request before the job fails.
            convert_dict_list_columns_to_json: bool
                If set to True, will convert any dict or list columns (which cannot by default be successfully loaded to BigQuery to JSON strings)
            **load_kwargs: kwargs
                Arguments to pass to the underlying load_table_from_uri call on the BigQuery
                client.
        """
        data_type = "csv"
        tmp_gcs_bucket = (
            tmp_gcs_bucket
            or self.tmp_gcs_bucket
            or check_env.check("GCS_TEMP_BUCKET", tmp_gcs_bucket)
        )
        if not tmp_gcs_bucket:
            raise ValueError(
                "Must set GCS_TEMP_BUCKET environment variable or pass in tmp_gcs_bucket parameter. If you have smaller data, you can use the `copy_direct` method to upload the data without needing to use CloudStorage. This alternate method will not work well with larger data."
            )

        if convert_dict_list_columns_to_json:
            tbl = self._stringify_records(tbl)

        job_config = self._prepare_local_upload_job(
            tbl,
            table_name,
            if_exists,
            max_errors,
            job_config,
            template_table,
            ignoreheader,
            nullas,
            allow_quoted_newlines,
            allow_jagged_rows,
            quote,
            schema,
        )

        gcs_client = gcs_client or GoogleCloudStorage(app_creds=self.app_creds)
        temp_blob_name = f"{uuid.uuid4()}.{data_type}"
        temp_blob_uri = gcs_client.upload_table(tbl, tmp_gcs_bucket, temp_blob_name)

        # load CSV from Cloud Storage into BigQuery
        try:
            self._load_table_from_uri(
                source_uris=temp_blob_uri,
                destination=self.get_table_ref(table_name=table_name),
                job_config=job_config,
                max_timeout=max_timeout,
                **load_kwargs,
            )
        finally:
            gcs_client.delete_blob(tmp_gcs_bucket, temp_blob_name)

    def _stringify_records(self, tbl):
        # Convert dict columns to JSON strings
        for field in tbl.get_columns_type_stats():
            if "dict" in field["type"] or "list" in field["type"]:
                new_petl = tbl.table.addfield(
                    field["name"] + "_replace",
                    lambda row, field=field: json.dumps(row[field["name"]]),
                )
                new_tbl = Table(new_petl)
                new_tbl.remove_column(field["name"])
                new_tbl.rename_column(field["name"] + "_replace", field["name"])
                new_tbl.materialize()
                tbl = new_tbl

        return tbl

    def _prepare_local_upload_job(
        self,
        tbl,
        table_name,
        if_exists,
        max_errors,
        job_config,
        template_table,
        ignoreheader,
        nullas,
        allow_quoted_newlines,
        allow_jagged_rows,
        quote,
        schema,
    ):
        data_type = "csv"

        self._validate_copy_inputs(if_exists=if_exists, data_type=data_type)

        # If our source table is loaded from CSV with no transformations
        # The original source file will be directly loaded to GCS
        # We may need to pass along a custom delimiter to BigQuery
        # Otherwise we use the default comma
        if isinstance(tbl.table, petl.io.csv_py3.CSVView):
            csv_delimiter = tbl.table.csvargs.get("delimiter", ",")
        else:
            csv_delimiter = ","

        job_config = self._process_job_config(
            job_config=job_config,
            destination_table_name=table_name,
            if_exists=if_exists,
            max_errors=max_errors,
            data_type=data_type,
            template_table=template_table,
            parsons_table=tbl,
            ignoreheader=ignoreheader,
            nullas=nullas,
            allow_quoted_newlines=allow_quoted_newlines,
            allow_jagged_rows=allow_jagged_rows,
            quote=quote,
            custom_schema=schema,
            csv_delimiter=csv_delimiter,
        )

        # Reorder schema to match table to ensure compatibility
        schema = []
        for column in tbl.columns:
            try:
                schema_row = [i for i in job_config.schema if i.name.lower() == column.lower()][0]
            except IndexError as e:
                raise IndexError(
                    f"Column found in Table that was not found in schema: {column}"
                ) from e
            schema.append(schema_row)
        job_config.schema = schema

        return job_config

    def duplicate_table(
        self,
        source_table,
        destination_table,
        if_exists="fail",
        drop_source_table=False,
    ):
        """
        Create a copy of an existing table (or subset of rows) in a new
        table.

        `Args:`
            source_table: str
                Name of existing schema and table (e.g. ``myschema.oldtable``)
            destination_table: str
                Name of destination schema and table (e.g. ``myschema.newtable``)
            if_exists: str
                If the table already exists, either ``fail``, ``replace``, or
                ``ignore`` the operation.
            drop_source_table: boolean
                Drop the source table
        """
        if if_exists not in ["fail", "replace", "ignore"]:
            raise ValueError("Invalid value for `if_exists` argument")
        if if_exists == "fail" and self.table_exists(destination_table):
            raise ValueError("Table already exists.")

        table__replace_clause = "OR REPLACE " if if_exists == "replace" else ""
        table__exists_clause = " IF NOT EXISTS" if if_exists == "ignore" else ""

        query = f"""
            CREATE {table__replace_clause}TABLE{table__exists_clause}
            {destination_table}
            CLONE {source_table}
        """
        self.query(sql=query, return_values=False)
        if drop_source_table:
            self.delete_table(table_name=source_table)

    def upsert(
        self,
        table_obj,
        target_table,
        primary_key,
        distinct_check=True,
        cleanup_temp_table=True,
        from_s3=False,
        **copy_args,
    ):
        r"""
        Preform an upsert on an existing table. An upsert is a function in which rows
        in a table are updated and inserted at the same time.

        `Args:`
            table_obj: obj
                A Parsons table object
            target_table: str
                The schema and table name to upsert
            primary_key: str or list
                The primary key column(s) of the target table
            distinct_check: boolean
                Check if the primary key column is distinct. Raise error if not.
            cleanup_temp_table: boolean
                A temp table is dropped by default on cleanup. You can set to False for debugging.
            from_s3: boolean
                Instead of specifying a table_obj (set the first argument to None),
                set this to True and include :func:`~parsons.databases.bigquery.Bigquery.copy_s3`
                arguments to upsert a pre-existing s3 file into the target_table
            **copy_args: kwargs
                See :func:`~parsons.databases.bigquery.BigQuery.copy` for options.
        """
        if not self.table_exists(target_table):
            logger.info(
                "Target table does not exist. Copying into newly \
                         created target table."
            )

            self.copy(table_obj, target_table)
            return None

        if isinstance(primary_key, str):
            primary_keys = [primary_key]
        else:
            primary_keys = primary_key

        if distinct_check:
            primary_keys_statement = ", ".join(primary_keys)
            diff = self.query(
                f"""
                select (
                    select count(*)
                    from {target_table}
                ) - (
                    SELECT COUNT(*) from (
                        select distinct {primary_keys_statement}
                        from {target_table}
                    )
                ) as total_count
            """
            ).first
            if diff > 0:
                raise ValueError("Primary key column contains duplicate values.")

        noise = f"{random.randrange(0, 10000):04}"[:4]
        date_stamp = datetime.datetime.now().strftime("%Y%m%d_%H%M")
        # Generate a temp table like "table_tmp_20200210_1230_14212"
        staging_tbl = f"{target_table}_stg_{date_stamp}_{noise}"

        # Copy to a staging table
        logger.info(f"Building staging table: {staging_tbl}")

        if from_s3:
            if table_obj is not None:
                raise ValueError(
                    "upsert(... from_s3=True) requires the first argument (table_obj)"
                    " to be None. from_s3 and table_obj are mutually exclusive."
                )
            self.copy_s3(staging_tbl, template_table=target_table, **copy_args)

        else:
            self.copy(
                tbl=table_obj,
                table_name=staging_tbl,
                template_table=target_table,
                **copy_args,
            )

        # Delete rows
        comparisons = [
            f"`{staging_tbl}`.{primary_key} = `{target_table}`.{primary_key}"
            for primary_key in primary_keys
        ]
        where_clause = " and ".join(comparisons)

        queries = [
            f"""
                DELETE FROM `{target_table}`
                WHERE EXISTS
                (SELECT * FROM `{staging_tbl}`
                WHERE {where_clause})
                """,
            f"""
                INSERT INTO `{target_table}`
                SELECT * FROM `{staging_tbl}`
                """,
        ]

        try:
            return self.query_with_transaction(queries=queries)
        finally:
            if cleanup_temp_table:
                logger.info(f"Deleting staging table: {staging_tbl}")
                self.query(f"DROP TABLE IF EXISTS {staging_tbl}", return_values=False)

    def delete_table(self, table_name):
        """
        Delete a BigQuery table.

        `Args:`
            table_name: str
                The name of the table to delete.
        """
        table_ref = self.get_table_ref(table_name=table_name)
        self.client.delete_table(table_ref)

    def table_exists(self, table_name: str) -> bool:
        """
        Check whether or not the Google BigQuery table exists in the specified dataset.

        `Args:`
            table_name: str
                The name of the BigQuery table to check for
        `Returns:`
            bool
                True if the table exists in the specified dataset, false otherwise
        """
        table_ref = self.get_table_ref(table_name=table_name)
        try:
            self.client.get_table(table_ref)
        except exceptions.NotFound:
            return False

        return True

    def get_tables(self, schema, table_name: Optional[str] = None):
        """
        List the tables in a schema including metadata.

        Args:
            schema: str
                Filter by a schema
            table_name: str
                Filter by a table name
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.debug("Retrieving tables info.")
        sql = f"select * from {schema}.INFORMATION_SCHEMA.TABLES"
        if table_name:
            sql += f" where table_name = '{table_name}'"
        return self.query(sql)

    def get_views(self, schema, view: Optional[str] = None):
        """
        List views.

        Args:
            schema: str
                Filter by a schema
            view: str
                Filter by a table name
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.debug("Retrieving views info.")
        sql = f"""
              select
                table_schema as schema_name,
                table_name as view_name,
                view_definition
              from {schema}.INFORMATION_SCHEMA.VIEWS
              """
        if view:
            sql += f" where table_name = '{view}'"
        return self.query(sql)

    def get_columns(self, schema: str, table_name: str):
        """
        Gets the column names (and other column metadata) for a table. If you
        need just the column names run ``get_columns_list()``, as it is faster.

        `Args:`
            schema: str
                The schema name
            table_name: str
                The table name

        `Returns:`
            A dictionary mapping column name to a dictionary with extra info. The
            keys of the dictionary are ordered just liked the columns in the table.
            The extra info is a dict with format
        """

        base_query = f"""
        SELECT
            *
        FROM `{schema}.INFORMATION_SCHEMA.COLUMNS`
        WHERE
            table_name = '{table_name}'
        """

        logger.debug(base_query)

        return {
            row["column_name"]: {
                "data_type": row["data_type"],
                "is_nullable": row["is_nullable"],
                "is_updatable": row["is_updatable"],
                "is_partioning_column": row["is_partitioning_column"],
                "rounding_mode": row["rounding_mode"],
            }
            for row in self.query(base_query)
        }

    def get_columns_list(self, schema: str, table_name: str) -> list:
        """
        Gets the column names for a table.

        `Args:`
            schema: str
                The schema name
            table_name: str
                The table name

        `Returns:`
            A list of column names
        """

        table_ref = self.client.get_table(table=f"{schema}.{table_name}")

        return [schema_ref.name for schema_ref in table_ref.schema]

    def get_row_count(self, schema: str, table_name: str) -> int:
        """
        Gets the row count for a BigQuery materialization.

        Caution: This method uses SELECT COUNT(*) which can be expensive for large tables,
        especially those with many columns. This is because BigQuery scans all table data
        to perform the count, even though only the row count is returned.

        `Args`:
            schema: str
                The schema name
            table_name: str
                The table name

        `Returns:`
            Row count of the target table
        """

        sql = f"SELECT COUNT(*) AS row_count FROM `{schema}.{table_name}`"
        result = self.query(sql=sql)

        return result["row_count"][0]

    def get_table_ref(self, table_name):
        # Helper function to build a TableReference for our table
        parsed = parse_table_name(table_name)
        dataset_ref = self.client.dataset(parsed["dataset"])
        return dataset_ref.table(parsed["table"])

    def _get_job_config_schema(
        self,
        job_config: LoadJobConfig,
        destination_table_name: str,
        if_exists: str,
        parsons_table: Optional[Table] = None,
        custom_schema: Optional[list] = None,
        template_table: Optional[str] = None,
    ) -> Optional[List[bigquery.SchemaField]]:
        # if job.schema already set in job_config, do nothing
        if job_config.schema:
            return job_config.schema
        # if schema specified by user, convert to schema type and use that
        if custom_schema:
            return map_column_headers_to_schema_field(custom_schema)
        # if template_table specified by user, use that
        # otherwise, if loading into existing table, infer destination table as template table
        if not template_table and if_exists in ("append", "truncate"):
            template_table = destination_table_name
        # if template_table set, use it to set the load schema
        if template_table:
            try:
                bigquery_table = self.client.get_table(template_table)
                return bigquery_table.schema
            except google.api_core.exceptions.NotFound:
                logger.warning(
                    f"template_table '{template_table}' not found. Unable to set schema."
                )
        # if load is coming from a Parsons table, use that to generate schema
        if parsons_table:
            return self._generate_schema_from_parsons_table(parsons_table)

        return None

    def _generate_schema_from_parsons_table(self, tbl):
        """BigQuery schema generation based on contents of Parsons table.

        Not usually necessary to use this. BigQuery is able to
        natively autodetect schema formats."""
        stats = tbl.get_columns_type_stats()
        fields = []
        for stat in stats:
            petl_types = stat["type"]

            # Prefer 'str' if included
            # Otherwise choose first type that isn't "NoneType"
            # Otherwise choose NoneType
            not_none_petl_types = [i for i in petl_types if i != "NoneType"]
            if "str" in petl_types:
                best_type = "str"
            elif ("int" in petl_types) and ("float" in petl_types):
                best_type = "float"
            elif not_none_petl_types:
                best_type = not_none_petl_types[0]
            else:
                best_type = "NoneType"

            # Python datetimes may be datetime or timestamp in BigQuery
            # BigQuery datetimes have no timezone, timestamps do
            if best_type == "datetime":
                for value in petl.util.base.values(tbl.table, stat["name"]):
                    if isinstance(value, datetime.datetime) and value.tzinfo:
                        best_type = "timestamp"

            try:
                field_type = self._bigquery_type(best_type)
            except KeyError as e:
                raise KeyError(
                    "Column type not supported for load to BigQuery. "
                    "Consider converting to another type. "
                    f"[type={best_type}]"
                ) from e
            field = bigquery.schema.SchemaField(stat["name"], field_type)
            fields.append(field)
        return fields

    def _process_job_config(
        self,
        destination_table_name: str,
        if_exists: str,
        max_errors: int,
        data_type: str,
        csv_delimiter: Optional[str] = ",",
        ignoreheader: Optional[int] = 1,
        nullas: Optional[str] = None,
        allow_quoted_newlines: Optional[bool] = None,
        allow_jagged_rows: Optional[bool] = None,
        quote: Optional[str] = None,
        job_config: Optional[LoadJobConfig] = None,
        custom_schema: Optional[list] = None,
        template_table: Optional[str] = None,
        parsons_table: Optional[Table] = None,
    ) -> LoadJobConfig:
        """
        Internal function to neatly process a user-supplied job configuration object.
        As a convention, if both the job_config and keyword arguments specify a value,
        we defer to the job_config.

        `Args`:
            job_config: `LoadJobConfig`
                Optionally supplied GCS `LoadJobConfig` object

        `Returns`:
            A `LoadJobConfig` object
        """

        if not job_config:
            job_config = bigquery.LoadJobConfig()

        job_config.schema = self._get_job_config_schema(
            job_config=job_config,
            destination_table_name=destination_table_name,
            if_exists=if_exists,
            parsons_table=parsons_table,
            custom_schema=custom_schema,
            template_table=template_table,
        )
        if not job_config.schema:
            job_config.autodetect = True

        if not job_config.create_disposition:
            job_config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED

        if not job_config.max_bad_records:
            job_config.max_bad_records = max_errors

        if not job_config.skip_leading_rows and data_type == "csv":
            job_config.skip_leading_rows = ignoreheader

        if not job_config.source_format:
            job_config.source_format = (
                bigquery.SourceFormat.CSV
                if data_type == "csv"
                else bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
            )

        if not job_config.field_delimiter:
            if data_type == "csv":
                job_config.field_delimiter = csv_delimiter
            if nullas:
                job_config.null_marker = nullas

        destination_table_exists = self.table_exists(destination_table_name)
        if not job_config.write_disposition:
            if if_exists == "append":
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
            elif if_exists == "truncate":
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
            elif destination_table_exists and if_exists == "fail":
                raise Exception("Table already exists.")
            elif if_exists == "drop" and destination_table_exists:
                self.delete_table(destination_table_name)
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_EMPTY
            else:
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_EMPTY

        if not job_config.allow_quoted_newlines and allow_quoted_newlines is not None:
            job_config.allow_quoted_newlines = allow_quoted_newlines

        if data_type == "csv" and allow_jagged_rows is not None:
            job_config.allow_jagged_rows = allow_jagged_rows

        if not job_config.quote_character and quote is not None:
            job_config.quote_character = quote

        return job_config

    def _fetch_query_results(self, cursor) -> Table:
        # We will use a temp file to cache the results so that they are not all living
        # in memory. We'll use pickle to serialize the results to file in order to maintain
        # the proper data types (e.g. integer).
        temp_filename = create_temp_file()

        with open(temp_filename, "wb") as temp_file:
            header = [i[0] for i in cursor.description]
            pickle.dump(header, temp_file)

            while True:
                batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                if len(batch) == 0:
                    break

                for row in batch:
                    row_data = list(row.values())
                    pickle.dump(row_data, temp_file)

        ptable = petl.frompickle(temp_filename)
        return Table(ptable)

    def _validate_copy_inputs(self, if_exists: str, data_type: str):
        if if_exists not in ["fail", "truncate", "append", "drop"]:
            raise ValueError(
                f"Unexpected value for if_exists: {if_exists}, must be one of "
                '"append", "drop", "truncate", or "fail"'
            )
        if data_type not in ["csv", "json"]:
            raise ValueError(f"Only supports csv or json files [data_type = {data_type}]")

    def _load_table_from_uri(
        self, source_uris, destination, job_config, max_timeout, **load_kwargs
    ):
        load_job = self.client.load_table_from_uri(
            source_uris=source_uris,
            destination=destination,
            job_config=job_config,
            **load_kwargs,
        )

        try:
            load_job.result(timeout=max_timeout)
            return load_job
        except exceptions.BadRequest as e:
            for idx, error_ in enumerate(load_job.errors):
                if idx == 0:
                    logger.error("* Load job failed. Enumerating errors collection below:")
                logger.error(f"** Error collection - index {idx}:")
                logger.error(error_)

            raise e

    @staticmethod
    def _bigquery_type(tp):
        return BIGQUERY_TYPE_MAP[tp]

    def table(self, table_name):
        # Return a MySQL table object

        return BigQueryTable(self, table_name)

    def extract(
        self,
        dataset: str,
        table_name: str,
        gcs_bucket: str,
        gcs_blob_name: str,
        project: Optional[str] = None,
        gzip: bool = False,
        location: str = "US",
        destination_file_format: str = "CSV",
        field_delimiter: str = ",",
        compression: str = None,
        job_config: ExtractJobConfig = None,
        wait_for_job_to_complete: bool = True,
        **export_kwargs,
    ) -> None:
        """
        Extracts a BigQuery table to a Google Cloud Storage bucket.

        Args:
            dataset (str): The BigQuery dataset containing the table.
            table_name (str): The name of the table to extract.
            gcs_bucket (str): The GCS bucket where the table will be
              exported.
            gcs_blob_name (str): The name of the blob in the GCS
              bucket.
            project (Optional[str]): The Google Cloud project ID. If
              not provided, the default project of the client is used.
            gzip (bool): If True, the exported file will be compressed
              using GZIP. Defaults to False.
        """
        if not job_config:
            logger.info("Using default job config as none was provided...")
            job_config = ExtractJobConfig(
                destination_format=destination_file_format,
                compression=compression,
                field_delimiter=field_delimiter,
            )
        source = f"{dataset}.{table_name}"
        gs_destination = f"gs://{gcs_bucket}/{gcs_blob_name}"
        compression = "GZIP" if gzip else compression
        logger.info(f"Project (parsons): {project}")
        extract_job = self.client.extract_table(
            source=source,
            destination_uris=gs_destination,
            location=location,
            job_config=job_config,
            project=project,
            **export_kwargs,
        )
        if wait_for_job_to_complete:
            extract_job.result()
            logger.info(f"Finished exporting query result to {gs_destination}.")

        return extract_job

    def copy_between_projects(
        self,
        source_project,
        source_dataset,
        source_table,
        destination_project,
        destination_dataset,
        destination_table,
        if_dataset_not_exists="fail",
        if_table_exists="fail",
    ):
        """
        Copy a table from one project to another. Fails if the source or target project
            does not exist.
        If the target dataset does not exist, fhe flag if_dataset_not_exists controls behavior.
            It defaults to 'fail'; set it to 'create' if it's ok to create it.
        If the target table exists, the flag if_table_exists controls behavior.
            It defaults to 'fail'; set it to 'overwrite' if it's ok to overwrite an existing table.

        `Args`:
            source_project: str
                Name of source project
            source_dataset: str
                Name of source dataset
            source_table: str
                Name of source table
            destination_project: str
                Name of destination project
            destination_dataset: str
                Name of destination dataset
            destination_table: str
                Name of destination table
            if_dataset_not_exists: str
                Action if dataset doesn't exist {'fail','create'}
            if_table_exists: str
                Action if table exists {'fail', 'overwrite'}

        `Returns:`
            None
        """

        from google.cloud import bigquery
        from google.cloud.exceptions import NotFound

        destination_table_id = (
            destination_project + "." + destination_dataset + "." + destination_table
        )
        source_table_id = source_project + "." + source_dataset + "." + source_table
        dataset_id = destination_project + "." + destination_dataset

        # check if destination dataset exists
        try:
            self.client.get_dataset(dataset_id)  # Make an API request.
            # if it exists: continue; if not, check to see if it's ok to create it
        except NotFound:
            # if it doesn't exist: check if it's ok to create it
            if if_dataset_not_exists == "create":  # create a new dataset in the destination
                dataset = bigquery.Dataset(dataset_id)
                dataset = self.client.create_dataset(dataset, timeout=30)
            else:  # if it doesn't exist and it's not ok to create it, fail
                logger.error("BigQuery copy failed")
                logger.error(
                    f"Dataset {destination_dataset} does not exist and if_dataset_not_exists set to {if_dataset_not_exists}"
                )

        job_config = bigquery.CopyJobConfig()

        # check if destination table exists
        try:
            self.client.get_table(destination_table_id)
            if if_table_exists == "overwrite":  # if it exists
                job_config = bigquery.CopyJobConfig()
                job_config.write_disposition = "WRITE_TRUNCATE"
                job = self.client.copy_table(
                    source_table_id,
                    destination_table_id,
                    location="US",
                    job_config=job_config,
                )
                result = job.result()
            else:
                logger.error(
                    f"BigQuery copy failed, Table {destination_table} exists and if_table_exists set to {if_table_exists}"
                )

        except NotFound:
            # destination table doesn't exist, so we can create one
            job = self.client.copy_table(
                source_table_id,
                destination_table_id,
                location="US",
                job_config=job_config,
            )
            result = job.result()
            logger.info(result)


class BigQueryTable(BaseTable):
    """BigQuery table object."""

    def drop(self, cascade=False):
        """
        Drop the table.
        """

        self.db.delete_table(self.table)

    def truncate(self):
        """
        Truncate the table.
        """

        self.db.query(f"TRUNCATE TABLE {self.table}")



import unittest

import requests_mock

from parsons import NationBuilder as NB

from .fixtures import GET_PEOPLE_RESPONSE, PERSON_RESPONSE


class TestNationBuilder(unittest.TestCase):
    def test_client(self):
        nb = NB("test-slug", "test-token")
        self.assertEqual(nb.client.uri, "https://test-slug.nationbuilder.com/api/v1/")
        self.assertEqual(
            nb.client.headers,
            {
                "authorization": "Bearer test-token",
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )

    def test_get_uri_success(self):
        self.assertEqual(NB.get_uri("foo"), "https://foo.nationbuilder.com/api/v1")
        self.assertEqual(NB.get_uri("bar"), "https://bar.nationbuilder.com/api/v1")

    def test_get_uri_errors(self):
        values = ["", "  ", None, 1337, {}, []]

        for v in values:
            with self.assertRaises(ValueError):
                NB.get_uri(v)

    def test_get_auth_headers_success(self):
        self.assertEqual(NB.get_auth_headers("foo"), {"authorization": "Bearer foo"})
        self.assertEqual(NB.get_auth_headers("bar"), {"authorization": "Bearer bar"})

    def test_get_auth_headers_errors(self):
        values = ["", "  ", None, 1337, {}, []]

        for v in values:
            with self.assertRaises(ValueError):
                NB.get_auth_headers(v)

    def test_parse_next_params_success(self):
        n, t = NB.parse_next_params("/a/b/c?__nonce=foo&__token=bar")
        self.assertEqual(n, "foo")
        self.assertEqual(t, "bar")

    def test_get_next_params_errors(self):
        with self.assertRaises(ValueError):
            NB.parse_next_params("/a/b/c?baz=1")

        with self.assertRaises(ValueError):
            NB.parse_next_params("/a/b/c?__nonce=1")

        with self.assertRaises(ValueError):
            NB.parse_next_params("/a/b/c?__token=1")

    def test_make_next_url(self):
        self.assertEqual(
            NB.make_next_url("example.com", "bar", "baz"),
            "example.com?limit=100&__nonce=bar&__token=baz",
        )

    @requests_mock.Mocker()
    def test_get_people_handle_empty_response(self, m):
        nb = NB("test-slug", "test-token")
        m.get("https://test-slug.nationbuilder.com/api/v1/people", json={"results": []})
        table = nb.get_people()
        self.assertEqual(table.num_rows, 0)

    @requests_mock.Mocker()
    def test_get_people(self, m):
        nb = NB("test-slug", "test-token")
        m.get(
            "https://test-slug.nationbuilder.com/api/v1/people",
            json=GET_PEOPLE_RESPONSE,
        )
        table = nb.get_people()

        self.assertEqual(table.num_rows, 2)
        self.assertEqual(len(table.columns), 59)

        self.assertEqual(table[0]["first_name"], "Foo")
        self.assertEqual(table[0]["last_name"], "Bar")
        self.assertEqual(table[0]["email"], "foo@example.com")

    @requests_mock.Mocker()
    def test_get_people_with_next(self, m):
        """Make two requests and get the same data twice. This will exercise the while loop."""
        nb = NB("test-slug", "test-token")

        GET_PEOPLE_RESPONSE_WITH_NEXT = GET_PEOPLE_RESPONSE.copy()
        GET_PEOPLE_RESPONSE_WITH_NEXT["next"] = (
            "https://test-slug.nationbuilder.com/api/v1/people?limit=100&__nonce=bar&__token=baz"
        )

        m.get(
            "https://test-slug.nationbuilder.com/api/v1/people",
            json=GET_PEOPLE_RESPONSE_WITH_NEXT,
        )

        m.get(
            "https://test-slug.nationbuilder.com/api/v1/people?limit=100&__nonce=bar&__token=baz",
            json=GET_PEOPLE_RESPONSE,
        )

        table = nb.get_people()

        self.assertEqual(table.num_rows, 4)
        self.assertEqual(len(table.columns), 59)

        self.assertEqual(table[1]["first_name"], "Zoo")
        self.assertEqual(table[1]["last_name"], "Baz")
        self.assertEqual(table[1]["email"], "bar@example.com")

    def test_update_person_raises_with_bad_params(self):
        nb = NB("test-slug", "test-token")

        with self.assertRaises(ValueError):
            nb.update_person(None, {})

        with self.assertRaises(ValueError):
            nb.update_person(1, {})

        with self.assertRaises(ValueError):
            nb.update_person(" ", {})

        with self.assertRaises(ValueError):
            nb.update_person("1", None)

        with self.assertRaises(ValueError):
            nb.update_person("1", "bad value")

    @requests_mock.Mocker()
    def test_update_person(self, m):
        """Requests the correct URL, returns the correct data and doesn't raise exceptions."""
        nb = NB("test-slug", "test-token")

        m.put(
            "https://test-slug.nationbuilder.com/api/v1/people/1",
            json=PERSON_RESPONSE,
        )

        response = nb.update_person("1", {"tags": ["zoot", "boot"]})
        person = response["person"]

        self.assertEqual(person["id"], 1)
        self.assertEqual(person["first_name"], "Foo")
        self.assertEqual(person["last_name"], "Bar")
        self.assertEqual(person["email"], "foo@example.com")

    def test_upsert_person_raises_with_bad_params(self):
        nb = NB("test-slug", "test-token")

        with self.assertRaises(ValueError):
            nb.upsert_person({"tags": ["zoot", "boot"]})

    @requests_mock.Mocker()
    def test_upsert_person(self, m):
        """Requests the correct URL, returns the correct data and doesn't raise exceptions."""
        nb = NB("test-slug", "test-token")

        m.put(
            "https://test-slug.nationbuilder.com/api/v1/people/push",
            json=PERSON_RESPONSE,
        )

        created, response = nb.upsert_person({"email": "foo@example.com"})
        self.assertFalse(created)

        person = response["person"]

        self.assertEqual(person["id"], 1)
        self.assertEqual(person["first_name"], "Foo")
        self.assertEqual(person["last_name"], "Bar")
        self.assertEqual(person["email"], "foo@example.com")



import datetime
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import mock

import pytest

from parsons import Table
from parsons.utilities import check_env, files, json_format, sql_helpers
from parsons.utilities.datetime import date_to_timestamp, parse_date


@pytest.mark.parametrize(
    ["date", "exp_ts"],
    [
        pytest.param("2018-12-13", 1544659200),
        pytest.param("2018-12-13T00:00:00-08:00", 1544688000),
        pytest.param("", None),
        pytest.param(
            "2018-12-13 PST", None, marks=[pytest.mark.xfail(raises=ValueError, strict=True)]
        ),
    ],
)
def test_date_to_timestamp(date, exp_ts):
    assert date_to_timestamp(date) == exp_ts


def test_parse_date():
    # Test parsing an ISO8601 string
    expected = datetime.datetime(year=2020, month=1, day=1, tzinfo=datetime.timezone.utc)
    parsed = parse_date("2020-01-01T00:00:00.000 UTC")
    assert parsed == expected, parsed

    # Test parsing a unix timestamp
    parsed = parse_date(1577836800)
    assert parsed == expected, parsed

    # Test "parsing" a datetime object
    parsed = parse_date(expected)
    assert parsed == expected, parsed


#
# File utility tests (pytest-style)
#


def test_create_temp_file_for_path():
    temp_path = files.create_temp_file_for_path("some/file.gz")
    assert temp_path[-3:] == ".gz"


def test_create_temp_directory():
    temp_directory = files.create_temp_directory()
    test_file1 = f"{temp_directory}/test.txt"
    test_file2 = f"{temp_directory}/test2.txt"
    with open(test_file1, "w") as fh1, open(test_file2, "w") as fh2:
        fh1.write("TEST")
        fh2.write("TEST")

    assert files.has_data(test_file1)
    assert files.has_data(test_file2)

    files.cleanup_temp_directory(temp_directory)

    # Verify the temp file no longer exists
    with pytest.raises(FileNotFoundError):
        open(test_file1, "r")


def test_close_temp_file():
    temp = files.create_temp_file()
    files.close_temp_file(temp)

    # Verify the temp file no longer exists
    with pytest.raises(FileNotFoundError):
        open(temp, "r")


def test_is_gzip_path():
    assert files.is_gzip_path("some/file.gz")
    assert not files.is_gzip_path("some/file")
    assert not files.is_gzip_path("some/file.csv")


def test_suffix_for_compression_type():
    assert files.suffix_for_compression_type(None) == ""
    assert files.suffix_for_compression_type("") == ""
    assert files.suffix_for_compression_type("gzip") == ".gz"


def test_compression_type_for_path():
    assert files.compression_type_for_path("some/file") is None
    assert files.compression_type_for_path("some/file.csv") is None
    assert files.compression_type_for_path("some/file.csv.gz") == "gzip"


def test_empty_file():
    # Create fake files.
    tmp_folder = tempfile.mkdtemp()
    with open(Path(tmp_folder) / "empty.csv", "w+") as _:
        pass

    Table([["1"], ["a"]]).to_csv(str(Path(tmp_folder) / "full.csv"))

    assert not files.has_data(str(Path(tmp_folder) / "empty.csv"))
    assert files.has_data(str(Path(tmp_folder) / "full.csv"))

    # Remove fake files and dir
    shutil.rmtree(tmp_folder)


def test_json_format():
    assert json_format.arg_format("my_arg") == "myArg"


def test_remove_empty_keys():
    # Assert key removed when None
    test_dict = {"a": None, "b": 2}
    assert json_format.remove_empty_keys(test_dict) == {"b": 2}

    # Assert key not removed when None
    test_dict = {"a": 1, "b": 2}
    assert json_format.remove_empty_keys(test_dict) == {"a": 1, "b": 2}

    # Assert that a nested empty string is removed
    test_dict = {"a": "", "b": 2}
    assert json_format.remove_empty_keys(test_dict) == {"b": 2}


def test_redact_credentials():
    # Test with quotes, escape characters, and line breaks
    test_str = """COPY schema.tablename
    FROM 's3://bucket/path/to/file.csv'
    credentials  'aws_access_key_id=string-\\'escaped-quote;
    aws_secret_access_key='string-escape-char\\\\'
    MANIFEST"""

    test_result = """COPY schema.tablename
    FROM 's3://bucket/path/to/file.csv'
    CREDENTIALS REDACTED
    MANIFEST"""

    assert sql_helpers.redact_credentials(test_str) == test_result


class TestCheckEnv(unittest.TestCase):
    def test_environment_field(self):
        """Test check field"""
        result = check_env.check("PARAM", "param")
        self.assertEqual(result, "param")

    @mock.patch.dict(os.environ, {"PARAM": "env_param"})
    def test_environment_env(self):
        """Test check env"""
        result = check_env.check("PARAM", None)
        self.assertEqual(result, "env_param")

    @mock.patch.dict(os.environ, {"PARAM": "env_param"})
    def test_environment_field_env(self):
        """Test check field with env and field"""
        result = check_env.check("PARAM", "param")
        self.assertEqual(result, "param")

    def test_envrionment_error(self):
        """Test check env raises error"""
        with self.assertRaises(KeyError) as _:
            check_env.check("PARAM", None)
